"""The spectral model bases."""

from __future__ import annotations

import inspect
import warnings
from abc import ABC, ABCMeta, abstractmethod
from collections.abc import Mapping, Sequence
from enum import Enum
from functools import reduce
from typing import TYPE_CHECKING, NamedTuple

import astropy.units as u
import jax
import jax.numpy as jnp
import numpy as np
from astropy.cosmology import Planck18
from scipy.sparse import sparray

from elisa.data.base import ObservationData, ResponseData, SpectrumData
from elisa.models.parameter import Parameter, UniformParameter
from elisa.util.misc import build_namespace, define_fdjvp, make_pretty_table

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Literal

    from astropy.cosmology import LambdaCDM
    from numpyro.distributions import Distribution

    from elisa.models.parameter import ParamInfo
    from elisa.util.integrate import IntegralFactory
    from elisa.util.typing import (
        AdditiveFn,
        ArrayLike,
        CompEval,
        CompID,
        CompIDParamValMapping,
        CompIDStrMapping,
        CompName,
        CompParamName,
        ConvolveEval,
        JAXArray,
        JAXFloat,
        ModelCompiledFn,
        ModelEval,
        NameValMapping,
        ParamID,
        ParamIDStrMapping,
        ParamIDValMapping,
        ParamNameValMapping,
    )

    NDArray = np.ndarray


class Model(ABC):
    """Base model class."""

    _comps: tuple[Component, ...]
    __initialized: bool = False

    def __init__(
        self,
        name: str,
        latex: str,
        comps: list[Component],
    ):
        self._name = str(name)
        self._latex = str(latex)
        self._comps = tuple(comps)
        self._comps_id = tuple(comp._id for comp in comps)

        cname = build_namespace([c.name for c in comps])['namespace']
        cid_to_cname = dict(zip(self._comps_id, cname, strict=True))
        self._cid_to_cname = cid_to_cname

        self.__name = self._id_to_label(cid_to_cname, 'name')

        for name, comp in zip(cid_to_cname.values(), comps, strict=True):
            setattr(self, name, comp)

        self.__initialized = True

    @property
    def _cid_to_clatex(self) -> CompIDStrMapping:
        clatex = [c.latex for c in self._comps]
        clatex = build_namespace(clatex, latex=True)['namespace']
        return dict(zip(self._comps_id, clatex, strict=True))

    def compile(self, *, model_info: ModelInfo | None = None) -> CompiledModel:
        """Compile the model for fast evaluation.

        Parameters
        ----------
        model_info : ModelInfo, optional
            Optional model information used to compile the model.

        Returns
        -------
        CompiledModel
            The compiled model.
        """
        if self.type == 'conv':
            raise RuntimeError(f'cannot compile convolution model {self.name}')

        if model_info is None:
            model_info = get_model_info(
                self._comps, self._cid_to_cname, self._cid_to_clatex
            )
        else:
            if not isinstance(model_info, ModelInfo):
                raise TypeError('`model_info` must be a ModelInfo instance')

            if not set(self._comps_id).issubset(set(model_info.cid_to_params)):
                raise ValueError('inconsistent model information')

        # model name
        name = self._id_to_label(model_info.cid_to_name, 'name')

        # model parameter id
        params_id = []
        fixed_id = []
        for c in self._comps:
            for p in c.param_names:
                for pid in c[p]._nodes_id:
                    if pid in model_info.sample:
                        params_id.append(pid)
                    elif pid in model_info.fixed:
                        fixed_id.append(pid)

        # compiled model evaluation function
        fn = self._compile_model_fn(model_info)
        additive_fn = self._compile_additive_fn(model_info)

        # model type
        mtype = self.type

        return CompiledModel(
            name, params_id, fixed_id, fn, additive_fn, mtype, model_info
        )

    @property
    @abstractmethod
    def eval(self) -> ModelEval:
        """Get side-effect free model evaluation function."""
        pass

    @property
    def name(self) -> str:
        """Model name."""
        return self.__name

    @property
    def latex(self) -> str:
        r""":math:`\LaTeX` format of the model."""
        latex = [c.latex for c in self._comps]
        cid_to_latex = dict(
            zip(
                self._comps_id,
                build_namespace(latex, latex=True)['namespace'],
                strict=True,
            )
        )

        return self._id_to_label(cid_to_latex, 'latex')

    @property
    @abstractmethod
    def type(self) -> Literal['add', 'mul', 'conv']:
        """Model type."""
        pass

    @property
    def comp_names(self) -> tuple[CompName, ...]:
        """Component names."""
        return tuple(self._cid_to_cname.values())

    @property
    def _additive_comps(self) -> tuple[Model, ...]:
        if self.type != 'add':
            return ()
        else:
            return (self,)

    def _id_to_label(
        self,
        mapping: CompIDStrMapping,
        label_type: Literal['name', 'latex'],
    ) -> str:
        """Get the label of the parameter."""
        if label_type not in {'name', 'latex'}:
            raise ValueError(f'unknown label type: {label_type}')

        label = self._name if label_type == 'name' else self._latex

        for k, v in mapping.items():
            label = label.replace(k, v)

        return label

    def _compile_model_fn(self, model_info: ModelInfo) -> ModelCompiledFn:
        """Get the model evaluation function."""
        pid_to_value = {
            c._id: model_info.cid_to_params[c._id] for c in self._comps
        }

        eval_fn = jax.jit(self.eval)

        @jax.jit
        def fn(egrid: JAXArray, params: ParamIDValMapping) -> JAXArray:
            """The model evaluation function"""
            comps_params = jax.tree.map(lambda f: f(params), pid_to_value)
            return eval_fn(egrid, comps_params)

        for integrate in model_info.integrate.values():
            fn = jax.jit(integrate(fn))

        return fn

    def _compile_additive_fn(self, model_info: ModelInfo) -> AdditiveFn | None:
        """Get the additive model evaluation function."""
        additive_comps = self._additive_comps

        if len(additive_comps) <= 1:
            return None

        fns = [comp._compile_model_fn(model_info) for comp in additive_comps]
        comps_labels = [
            comp._id_to_label(model_info.cid_to_latex, 'latex')
            for comp in additive_comps
        ]

        @jax.jit
        def _fn(egrid: JAXArray, params: ParamIDValMapping) -> JAXArray:
            return [f(egrid, params) for f in fns]

        @jax.jit
        def fn(
            egrid: JAXArray, params: ParamIDValMapping
        ) -> dict[tuple[str, str], JAXArray]:
            """The evaluation function of additive components."""
            return dict(zip(comps_labels, _fn(egrid, params), strict=True))

        return fn

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        fields = ['No.', 'Component', 'Parameter', 'Value', 'Bound', 'Prior']
        info = get_model_info(
            self._comps, self._cid_to_cname, self._cid_to_clatex
        ).info
        tab_params = make_pretty_table(fields, info)
        mtype = get_mtype_str(self)
        return f'{mtype} Model: {self.name}\n{tab_params.get_string()}'

    def _repr_html_(self) -> str:
        fields = ['No.', 'Component', 'Parameter', 'Value', 'Bound', 'Prior']
        info = get_model_info(
            self._comps, self._cid_to_cname, self._cid_to_clatex
        ).info
        tab_params = make_pretty_table(fields, info)
        mtype = get_mtype_str(self)
        return (
            '<details open>'
            f'<summary><b>{mtype} Model: {self.name}</b></summary>'
            f'{tab_params.get_html_string(format=True)}'
            '</details>'
        )

    def __delattr__(self, key: str):
        if self.__initialized and hasattr(self, key):
            raise AttributeError("can't delete attribute")

        super().__delattr__(key)

    def __setattr__(self, key: str, value: Any):
        if self.__initialized and (
            not hasattr(self, key) or key in self._cid_to_cname.values()
        ):
            raise AttributeError("can't set attribute")

        super().__setattr__(key, value)

    def __getitem__(self, key: str) -> Component:
        if key not in self._cid_to_cname.values():
            raise KeyError(key)

        return getattr(self, key)

    def __add__(self, other: Model) -> CompositeModel:
        return CompositeModel(self, other, '+')

    def __radd__(self, other: Model) -> CompositeModel:
        return CompositeModel(other, self, '+')

    def __mul__(self, other: Model) -> CompositeModel:
        return CompositeModel(self, other, '*')

    def __rmul__(self, other: Model) -> CompositeModel:
        return CompositeModel(other, self, '*')


class CompiledModel:
    """Model with fast evaluation and fixed configuration."""

    __initialized: bool = False

    def __init__(
        self,
        name: str,
        params_id: Sequence[ParamID],
        fixed_id: Sequence[ParamID],
        fn: ModelCompiledFn,
        additive_fn: AdditiveFn | None,
        mtype: Literal['add', 'mul'],
        model_info: ModelInfo,
    ):
        pname_to_pid = {model_info.name[pid]: pid for pid in params_id}
        self.name = name
        self._pname_to_pid = pname_to_pid
        self._params_name = tuple(pname_to_pid)
        self._params_id = params_id
        self._params_default = dict(model_info.default)
        self._value_sequence_to_params: Callable[
            [Sequence[JAXFloat]], ParamIDValMapping
        ] = jax.jit(
            lambda sequence: dict(zip(params_id, sequence, strict=True))
        )
        fixed_name_to_pid = {model_info.name[pid]: pid for pid in fixed_id}
        pname_pid_all = pname_to_pid | fixed_name_to_pid
        self._value_mapping_to_params: Callable[
            [ParamNameValMapping], ParamIDValMapping
        ] = jax.jit(
            lambda mapping: {
                v: mapping[k] for k, v in pname_pid_all.items() if k in mapping
            }
        )
        self._fn = fn
        self._additive_fn = additive_fn
        self._type = mtype
        self._model_info = model_info
        self._nparam = len(pname_to_pid)
        self.__initialized = True

    @property
    def params_name(self) -> tuple[str, ...]:
        """Parameter names."""
        return self._params_name

    @property
    def type(self) -> Literal['add', 'mul']:
        """Model type."""
        return self._type

    @property
    def has_comps(self) -> bool:
        """Whether the model has additive subcomponents."""
        return self._additive_fn is not None

    def _prepare_eval(self, params: ArrayLike | Sequence | Mapping | None):
        """Check if `params` is valid for the model."""
        if isinstance(params, np.ndarray | jax.Array | Sequence):
            params = jnp.atleast_1d(jnp.asarray(params, float))
            if params.shape[-1] != self._nparam:
                raise ValueError(
                    f'expected params shape (..., {self._nparam}), got '
                    f'{jnp.shape(params)}'
                )
            params = jnp.moveaxis(params, -1, 0)
            params = self._value_sequence_to_params(params)

        elif isinstance(params, Mapping):
            # if not set(self.params_name).issubset(params):
            #     missing = set(self.params_name) - set(params)
            #     raise ValueError(f'missing parameters: {", ".join(missing)}')

            params = jax.tree.map(jnp.asarray, params)
            params = self._value_mapping_to_params(params)

        elif params is None:
            params = self._params_default

        else:
            raise TypeError('params must be a array, sequence or mapping')

        fn = self._fn
        add_fn = self._additive_fn
        shapes = jax.tree.flatten(
            tree=jax.tree.map(jnp.shape, params),
            is_leaf=lambda i: isinstance(i, tuple),
        )[0]

        if self.params_name:
            if not shapes:
                raise ValueError('params are empty')

            shape = shapes[0]
            if any(s != shape for s in shapes[1:]):
                raise ValueError('all params must have the same shape')

            # iteratively vmap and jit over params dimensions
            # use the nested-jit trick to reduce the compilation time
            for _ in range(len(shape)):
                fn = jax.jit(jax.vmap(fn, in_axes=(None, 0)))

            if add_fn is not None:
                for _ in range(len(shape)):
                    add_fn = jax.jit(jax.vmap(add_fn, in_axes=(None, 0)))

        return fn, add_fn, params

    def eval(
        self,
        egrid: ArrayLike,
        params: Sequence | Mapping | None = None,
    ) -> JAXArray:
        """Evaluate the model.

        Parameters
        ----------
        egrid : ndarray
            Photon energy grid in units of keV.
        params : sequence or mapping, optional
            Parameter sequence or mapping for the model.

        Returns
        -------
        jax.Array
            The model value.
        """
        f, _, params = self._prepare_eval(params)

        return f(jnp.asarray(egrid, float), params)

    def ne(
        self,
        egrid: ArrayLike,
        params: Sequence | Mapping | None = None,
        comps: bool = False,
    ) -> JAXArray | dict[str, JAXArray]:
        r"""Calculate :math:`N(E)`.

        Parameters
        ----------
        egrid : ndarray
            Photon energy grid in units of keV.
        params : sequence or mapping, optional
            Parameter sequence or mapping for the model.
        comps : bool, optional
            Whether to return the result of each component. Defaults to False.

        Returns
        -------
        jax.Array, or dict[str, jax.Array]
            The differential photon flux in units of ph cm⁻² s⁻¹ keV⁻¹.
        """
        if self.type != 'add':
            msg = f'N(E) is undefined for {self.type} type model "{self}"'
            raise TypeError(msg)

        if comps and not self.has_comps:
            raise RuntimeError(f'{self} has no sub-models with additive type')

        egrid = jnp.asarray(egrid, float)
        de = egrid[1:] - egrid[:-1]

        if comps:
            _, additive_fn, params = self._prepare_eval(params)
            comps_value = additive_fn(egrid, params)
            ne = jax.tree.map(lambda v: v / de, comps_value)
        else:
            ne = self.eval(egrid, params) / de

        return ne

    def ene(
        self,
        egrid: ArrayLike,
        params: Sequence | Mapping | None = None,
        comps: bool = False,
    ) -> JAXArray | dict[str, JAXArray]:
        r"""Calculate :math:`E N(E)`, i.e. :math:`F(\nu)`.

        Parameters
        ----------
        egrid : ndarray
            Photon energy grid in units of keV.
        params : sequence or mapping, optional
            Parameter sequence or mapping for the model.
        comps : bool, optional
            Whether to return the result of each component. Defaults to False.

        Returns
        -------
        jax.Array, or dict[str, jax.Array]
            The differential energy flux in units of erg cm⁻² s⁻¹ keV⁻¹.
        """
        if self.type != 'add':
            msg = f'EN(E) is undefined for {self.type} type model "{self}"'
            raise TypeError(msg)

        if comps and not self.has_comps:
            raise RuntimeError(f'{self} has no sub-models with additive type')

        keV_to_erg = 1.602176634e-9
        egrid = jnp.asarray(egrid, float)
        emid = jnp.sqrt(egrid[:-1] * egrid[1:])
        factor = keV_to_erg * emid

        ne = self.ne(egrid, params, comps)

        if comps:
            ene = jax.tree.map(lambda v: factor * v, ne)
        else:
            ene = factor * ne

        return ene

    def eene(
        self,
        egrid: ArrayLike,
        params: Sequence | Mapping | None = None,
        comps: bool = False,
    ) -> JAXArray | dict[str, JAXArray]:
        r"""Calculate :math:`E^2 N(E)`, i.e. :math:`\nu F(\nu)`.

        Parameters
        ----------
        egrid : ndarray
            Photon energy grid in units of keV.
        params : sequence or mapping, optional
            Parameter sequence or mapping for the model.
        comps : bool, optional
            Whether to return the result of each component. Defaults to False.

        Returns
        -------
        jax.Array, or dict[str, jax.Array]
            The energy flux in units of erg cm⁻² s⁻¹.
        """
        if self.type != 'add':
            msg = f'EEN(E) is undefined for {self.type} type model "{self}"'
            raise TypeError(msg)

        if self._additive_fn is None and comps:
            raise RuntimeError(f'{self} has no sub-models with additive type')

        keV_to_erg = 1.602176634e-9
        egrid = jnp.asarray(egrid, float)
        e2 = egrid[:-1] * egrid[1:]
        factor = keV_to_erg * e2

        ne = self.ne(egrid, params, comps)

        if comps:
            eene = jax.tree.map(lambda v: factor * v, ne)
        else:
            eene = factor * ne

        return eene

    def ce(
        self,
        egrid: ArrayLike,
        resp_matrix: ArrayLike,
        channel_width: ArrayLike,
        params: Sequence | Mapping | None = None,
        comps: bool = False,
    ) -> JAXArray | dict[str, JAXArray]:
        r"""Calculate the folded model, i.e. :math:`C(E)`.

        Parameters
        ----------
        egrid : ndarray
            Photon energy grid of `resp_matrix`, in units of keV.
        resp_matrix : ndarray
            Instrumental response matrix used to fold the model.
        channel_width : ndarray
            Measured energy channel width of `resp_matrix`.
        params : sequence or mapping, optional
            Parameter sequence or mapping for the model.
        comps : bool, optional
            Whether to return the result of each component. Defaults to False.

        Returns
        -------
        jax.Array, or dict[str, jax.Array]
            The folded model in units of counts s⁻¹ keV⁻¹.
        """
        if self.type != 'add':
            msg = f'C(E) is undefined for {self.type} type model "{self}"'
            raise TypeError(msg)

        if comps and not self.has_comps:
            raise RuntimeError(f'{self} has no sub-models with additive type')

        egrid = jnp.asarray(egrid, float)
        resp_matrix = jnp.asarray(resp_matrix, float)
        channel_width = jnp.asarray(channel_width, float)

        ne = self.ne(egrid, params, comps)
        de = egrid[1:] - egrid[:-1]
        fn = jax.jit(lambda v: (v * de) @ resp_matrix / channel_width)

        if comps:
            return jax.tree.map(fn, ne)
        else:
            return fn(ne)

    def flux(
        self,
        emin: float | int | JAXFloat,
        emax: float | int | JAXFloat,
        params: Sequence | Mapping | None = None,
        energy: bool = True,
        comps: bool = False,
        ngrid: int = 1000,
        log: bool = True,
    ) -> jax.Array | dict[str, jax.Array]:
        r"""Calculate the flux of model between `emin` and `emax`.

        .. warning::
            The flux is calculated by trapezoidal rule, which may not be
            accurate if not enough energy bins are used when the difference
            between `emin` and `emax` is large.

        Parameters
        ----------
        emin : float or int
            Minimum value of energy range, in units of keV.
        emax : float or int
            Maximum value of energy range, in units of keV.
        params : sequence or mapping, optional
            Parameter sequence or mapping for the model.
        energy : bool, optional
            When True, calculate energy flux in units of erg cm⁻² s⁻¹;
            otherwise calculate photon flux in units of ph cm⁻² s⁻¹.
            The default is True.
        comps : bool, optional
            Whether to return the result of each component. The default is
            False.
        ngrid : int, optional
            The energy grid number to use in integration. The default is 1000.
        log : bool, optional
            Whether to use logarithmically regular energy grid. The default is
            True.

        Returns
        -------
        jax.Array, or dict[str, jax.Array]
            The model flux.
        """
        if self.type != 'add':
            msg = f'flux is undefined for {self.type} type model "{self}"'
            raise TypeError(msg)

        if log:
            egrid = jnp.geomspace(emin, emax, ngrid)
        else:
            egrid = jnp.linspace(emin, emax, ngrid)

        if energy:
            f = self.ene(egrid, params, comps)
        else:
            f = self.ne(egrid, params, comps)

        de = jnp.diff(egrid)
        fn = jax.jit(lambda v: jnp.sum(v * de, axis=-1))

        if comps:
            return jax.tree.map(fn, f)
        else:
            return fn(f)

    def lumin(
        self,
        emin_rest: float | int,
        emax_rest: float | int,
        z: float | int,
        params: Sequence | Mapping | None = None,
        comps: bool = False,
        ngrid: int = 1000,
        log: bool = True,
        cosmo: LambdaCDM = Planck18,
    ) -> JAXArray | dict[str, JAXArray]:
        """Calculate the luminosity of model.

        .. warning::
            The luminosity is calculated by trapezoidal rule, and is accurate
            only if enough numbers of energy grids are used.

        Parameters
        ----------
        emin_rest : float or int
            Minimum value of rest-frame energy range, in units of keV.
        emax_rest : float or int
            Maximum value of rest-frame energy range, in units of keV.
        z : float or int
            Redshift of the source.
        params : sequence or mapping, optional
            Parameter sequence or mapping for the model.
        comps : bool, optional
            Whether to return the result of each component. The default is
            False.
        ngrid : int, optional
            The energy grid number to use in integration. The default is 1000.
        log : bool, optional
            Whether to use logarithmically regular energy grid. The default is
            True.
        cosmo : LambdaCDM, optional
            Cosmology model used to calculate luminosity. The default is
            Planck18.

        Returns
        -------
        jax.Array, or dict[str, jax.Array]
            The model luminosity.
        """
        if self.type != 'add':
            msg = f'lumin is undefined for {self.type} type model "{self}"'
            raise TypeError(msg)

        z = float(z)
        flux = self.flux(
            emin=float(emin_rest) / (1.0 + z),
            emax=float(emax_rest) / (1.0 + z),
            params=params,
            energy=True,
            comps=bool(comps),
            ngrid=int(ngrid),
            log=bool(log),
        )
        flux_unit = u.Unit('erg cm^-2 s^-1')

        factor = 4.0 * np.pi * cosmo.luminosity_distance(z) ** 2
        to_lumin = lambda x: (x * flux_unit * factor).to('erg s^-1')

        if comps:
            return jax.tree.map(to_lumin, flux)
        else:
            return to_lumin(flux)

    def eiso(
        self,
        emin_rest: float | int,
        emax_rest: float | int,
        z: float | int,
        duration: float | int,
        params: Sequence | Mapping | None = None,
        comps: bool = False,
        ngrid: int = 1000,
        log: bool = True,
        cosmo: LambdaCDM = Planck18,
    ) -> JAXArray | dict[str, JAXArray]:
        r"""Calculate the isotropic emission energy of model.

        .. warning::
            The :math:`E_\mathrm{iso}` is calculated by trapezoidal rule,
            and is accurate only if enough numbers of energy grids are used.

        Parameters
        ----------
        emin_rest : float or int
            Minimum value of rest-frame energy range, in units of keV.
        emax_rest : float or int
            Maximum value of rest-frame energy range, in units of keV.
        z : float or int
            Redshift of the source.
        duration : float or int
            Observed duration of the source, in units of second.
        comps : bool, optional
            Whether to return the result of each component. The default is
            False.
        ngrid : int, optional
            The energy grid number to use in integration. The default is
            1000.
        log : bool, optional
            Whether to use logarithmically regular energy grid. The default
            is True.
        params : dict, optional
            Parameters dict to overwrite the fitted parameters.
        cosmo : LambdaCDM, optional
            Cosmology model used to calculate luminosity. The default is
            Planck18.

        Returns
        -------
        jax.Array, or dict[str, jax.Array]
            The isotropic emission energy of the model.
        """
        if self.type != 'add':
            msg = f'eiso is undefined for {self.type} type model "{self}"'
            raise TypeError(msg)

        lumin = self.lumin(
            emin_rest, emax_rest, z, params, comps, ngrid, log, cosmo
        )

        # This includes correction for time dilation.
        factor = duration / (1 + z) * u.s
        to_eiso = lambda x: (x * factor).to('erg')

        if comps:
            return jax.tree.map(to_eiso, lumin)
        else:
            return to_eiso(lumin)

    def simulate_from_data(
        self,
        data: ObservationData,
        spec_exposure: float | None = None,
        back_exposure: float | None = None,
        params: Sequence | Mapping | None = None,
        seed: int = 42,
        **kwargs: dict,
    ) -> ObservationData:
        """Simulate observation based on the configuration of existing data.

        Parameters
        ----------
        data : ObservationData
            Observation data to read observation configuration.
        spec_exposure : float, optional
            Exposure time of the source. Defaults to the exposure time of
            `spec_data`.
        back_exposure : float, optional
            Exposure time of the background. Defaults to the exposure time of
            `back_data`.
        params : sequence or mapping, optional
            Parameter sequence or mapping for the model.
        seed : int, optional
            Random seed for the simulation. The default is 42.
        **kwargs : dict, optional
            Additional keyword arguments passed to :class:`ObservationData`.

        Returns
        -------
        ObservationData
            Simulated observation data.
        """
        if self.type != 'add':
            msg = f'cannot simulate data from {self.type} type model "{self}"'
            raise TypeError(msg)

        if not isinstance(data, ObservationData):
            raise TypeError('data must be an ObservationData instance')

        if spec_exposure is None:
            spec_exposure = data.spec_exposure
        else:
            spec_exposure = float(spec_exposure)

        if data.has_back:
            back_counts = data.back_data.counts
            back_errors = data.back_data.errors
            back_poisson = data.back_poisson
            back_area_scale = data.back_data.area_scale
            back_back_scale = data.back_data.back_scale
            if back_exposure is None:
                back_exposure = data.back_exposure
            else:
                back_exposure = float(back_exposure)
        else:
            back_counts = back_errors = back_exposure = back_poisson = None
            back_area_scale = back_back_scale = 1.0

        spec_data = data.spec_data
        resp_data = data.resp_data

        simulation = self.simulate(
            photon_egrid=resp_data.photon_egrid,
            channel_emin=resp_data.channel_emin,
            channel_emax=resp_data.channel_emax,
            response_matrix=resp_data.sparse_matrix,
            spec_exposure=spec_exposure,
            spec_poisson=spec_data.poisson,
            spec_errors=spec_data.errors,
            back_counts=back_counts,
            back_errors=back_errors,
            back_exposure=back_exposure,
            back_poisson=back_poisson,
            spec_area_scale=spec_data.area_scale,
            spec_back_scale=spec_data.back_scale,
            back_area_scale=back_area_scale,
            back_back_scale=back_back_scale,
            quality=np.where(data.good_quality, 0, 1),
            grouping=data.grouping,
            channel=resp_data.channel,
            channel_type=resp_data.channel_type,
            response_sparse=resp_data.sparse,
            params=params,
            name=data.name,
            seed=seed,
            **kwargs,
        )
        simulation.set_erange(data.erange)

        return simulation

    def simulate(
        self,
        photon_egrid: NDArray,
        channel_emin: NDArray,
        channel_emax: NDArray,
        response_matrix: NDArray | sparray,
        spec_exposure: float,
        spec_poisson: bool = True,
        spec_errors: NDArray | None = None,
        back_counts: NDArray | None = None,
        back_errors: NDArray | None = None,
        back_exposure: float | None = None,
        back_poisson: bool | None = None,
        spec_area_scale: float | NDArray = 1.0,
        spec_back_scale: float | NDArray = 1.0,
        back_area_scale: float | NDArray = 1.0,
        back_back_scale: float | NDArray = 1.0,
        quality: NDArray | None = None,
        grouping: NDArray | None = None,
        channel: NDArray | None = None,
        channel_type: str = 'Ch',
        response_sparse: bool = False,
        params: Sequence | Mapping | None = None,
        name: str = 'simulation',
        seed: int = 42,
        **kwargs: dict,
    ) -> ObservationData:
        """Simulate observation data.

        Parameters
        ----------
        photon_egrid : ndarray
            Photon energy grid in units of keV.
        channel_emin : ndarray
            Lower energy bounds of the detector channels.
        channel_emax : ndarray
            Upper energy bounds of the detector channels.
        response_matrix : ndarray
            Response matrix of the detector.
        spec_exposure : float
            Exposure time of the source.
        spec_poisson : bool, optional
            Whether the source spectrum is Poisson distributed. If false,
            `spec_errors` must be provided. The default is True.
        spec_errors : ndarray, optional
            Errors of the source spectrum. Must be provided if `spec_poisson`
            is False.
        back_counts : ndarray, optional
            Background counts in each channel.
        back_errors : ndarray, optional
            Errors of the background counts. Must be provided if `back_poisson`
            is False.
        back_exposure : float, optional
            Exposure time of the background.
        back_poisson : bool, optional
            Whether the background spectrum is Poisson distributed. If false,
            `back_errors` must be provided.
        spec_area_scale : float or ndarray, optional
            Area scale factor of the source. The default is 1.0.
        spec_back_scale : float or ndarray, optional
            Background scale factor of the source. The default is 1.0.
        back_area_scale : float or ndarray, optional
            Area scale factor of the background. The default is 1.0.
        back_back_scale : float or ndarray, optional
            Background scale factor of the background. The default is 1.0.
        quality : ndarray, optional
            Quality flags of the data.
        grouping : ndarray, optional
            Grouping flags of the data.
        channel : ndarray, optional
            Channel numbers.
        channel_type : str, optional
            Channel type. The default is 'Ch'.
        response_sparse : bool, optional
            Whether the response matrix is sparse. The default is False.
        params : sequence or mapping, optional
            Parameter sequence or mapping for the model.
        name : str, optional
            Name of the simulation data. The default is 'simulation'.
        seed : int, optional
            Random seed for simulation. The default is 42.
        **kwargs : dict, optional
            Additional keyword arguments passed to :class:`ObservationData`.

        Returns
        -------
        ObservationData
            Simulated observation data.
        """
        if self.type != 'add':
            msg = f'cannot simulate data from {self.type} type model "{self}"'
            raise TypeError(msg)

        if channel is None:
            channel = np.arange(len(channel_emin)).astype(str)

        rng = np.random.default_rng(int(seed))

        resp_data = ResponseData(
            photon_egrid=photon_egrid,
            channel_emin=channel_emin,
            channel_emax=channel_emax,
            response_matrix=response_matrix,
            channel=channel,
            channel_type=channel_type,
            sparse=response_sparse,
        )

        if not spec_poisson:
            if spec_errors is None:
                raise ValueError(
                    'spec_errors must be provided if spec_poisson is False'
                )
            else:
                spec_errors = np.asarray(spec_errors, float)
                if spec_errors.size != resp_data.channel_number:
                    raise ValueError(
                        f'spec_errors size ({np.size(spec_errors)}) must be '
                        f'channel number ({resp_data.channel_number})'
                    )

        if not (
            back_counts is None
            and back_exposure is None
            and back_poisson is None
            and back_errors is None
        ):
            if back_counts is None:
                raise ValueError('back_counts must be also provided')
            if back_exposure is None:
                raise ValueError('back_exposure must be also provided')
            if back_poisson is None:
                raise ValueError('back_poisson must be also provided')

            back_counts = np.asarray(back_counts, np.float64)
            if back_counts.size != resp_data.channel_number:
                raise ValueError(
                    f'back_counts size ({np.size(back_counts)}) must be '
                    f'channel number ({resp_data.channel_number})'
                )

            if spec_poisson or back_poisson:
                if np.any(back_counts < 0):
                    warnings.warn(
                        'negative background counts is set to 0 for '
                        'Poisson counts simulation',
                        Warning,
                        stacklevel=2,
                    )
                    back_counts = np.clip(back_counts, 0, None)

            if not back_poisson:
                if back_errors is None:
                    raise ValueError(
                        'back_errors must be provided if back_poisson is False'
                    )
                else:
                    back_errors = np.asarray(back_errors, float)
                    if back_errors.size != resp_data.channel_number:
                        raise ValueError(
                            f'back_errors size ({np.size(back_errors)}) must '
                            f'be channel number ({resp_data.channel_number})'
                        )
            has_back = True
        else:
            has_back = False

        if has_back:
            if back_poisson:
                back_sim = np.array(rng.poisson(back_counts), np.float64)
                back_errors = np.sqrt(back_sim)
            else:
                # TODO: should fractional systematic errors be added?
                back_sim = rng.normal(back_counts, back_errors)

            back_data = SpectrumData(
                counts=back_sim,
                errors=back_errors,
                poisson=back_poisson,
                exposure=back_exposure,
                quality=quality,
                grouping=grouping,
                area_scale=back_area_scale,
                back_scale=back_back_scale,
            )
        else:
            back_data = None

        matrix = resp_data.sparse_matrix.T
        folded_rate = matrix @ self.eval(resp_data.photon_egrid, params)
        spec_counts = folded_rate * spec_exposure * spec_area_scale

        if has_back:
            back_ratio = (
                spec_exposure * spec_area_scale * spec_back_scale
            ) / (back_exposure * back_area_scale * back_back_scale)
            spec_counts += back_ratio * back_counts

        if spec_poisson:
            spec_sim = np.array(rng.poisson(spec_counts), np.float64)
            spec_errors = np.sqrt(spec_sim)
        else:
            # TODO: should fractional systematic errors be added?
            spec_sim = rng.normal(spec_counts, spec_errors)

        spec_data = SpectrumData(
            counts=spec_sim,
            errors=spec_errors,
            poisson=spec_poisson,
            exposure=spec_exposure,
            quality=quality,
            grouping=grouping,
            area_scale=spec_area_scale,
            back_scale=spec_back_scale,
        )

        return ObservationData(
            name=name,
            erange=[resp_data.channel_emin[0], resp_data.channel_emax[-1]],
            spec_data=spec_data,
            resp_data=resp_data,
            back_data=back_data,
            **kwargs,
        )

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        fields = ['No.', 'Component', 'Parameter', 'Value', 'Bound', 'Prior']
        info = self._model_info.info
        tab_params = make_pretty_table(fields, info)
        mtype = get_mtype_str(self)
        return f'{mtype} Model: {self.name}\n{tab_params.get_string()}'

    def _repr_html_(self) -> str:
        fields = ['No.', 'Component', 'Parameter', 'Value', 'Bound', 'Prior']
        info = self._model_info.info
        tab_params = make_pretty_table(fields, info)
        mtype = get_mtype_str(self)
        return (
            '<details open>'
            f'<summary><b>{mtype} Model: {self.name}</b></summary>'
            f'{tab_params.get_html_string(format=True)}'
            '</details>'
        )

    def __delattr__(self, key: str):
        if self.__initialized and hasattr(self, key):
            raise AttributeError("can't delete attribute")

        super().__delattr__(key)

    def __setattr__(self, key: str, value: Any):
        if self.__initialized:
            raise AttributeError("can't set attribute")

        super().__setattr__(key, value)


class UniComponentModel(Model):
    """Model defined by a single additive or multiplicative component."""

    _component: Component

    def __init__(self, component: Component):
        self._component = component
        super().__init__(component._id, component._id, [component])

    @property
    def eval(self) -> ModelEval:
        comp_id = self._component._id
        _fn = self._component.eval

        def fn(egrid: JAXArray, params: CompIDParamValMapping) -> JAXArray:
            """The model evaluation function"""
            return _fn(egrid, params[comp_id])

        return jax.jit(fn)

    @property
    def type(self) -> Literal['add', 'mul']:
        return self._component.type


class CompositeModel(Model):
    """Model defined by sum or product of two models."""

    _operands: tuple[Model, Model]
    _op: Callable[[JAXArray, JAXArray], JAXArray]
    _op_symbol: Literal['+', '*']
    _type: Literal['add', 'mul']
    __additive_comps: tuple[Model, ...] | None = None

    def __init__(self, lhs: Model, rhs: Model, op: Literal['+', '*']):
        # check if the type of lhs and rhs are both model
        if not (isinstance(lhs, Model) and isinstance(rhs, Model)):
            raise TypeError(
                f'unsupported operand types for {op}: '
                f"'{type(lhs).__name__}' and '{type(rhs).__name__}'"
            )

        self._operands = (lhs, rhs)
        self._op_symbol = op

        # check if operand is legal for the operator
        type1 = lhs.type
        type2 = rhs.type

        if op == '+':
            if type1 != 'add':
                raise TypeError(f'{lhs} is not an additive model')
            if type2 != 'add':
                raise TypeError(f'{rhs} is not an additive model')

            self._type = 'add'
            self._op = jnp.add
            op_name = '{} + {}'
            op_latex = '{{{}}} + {{{}}}'

        elif op == '*':
            if 'add' == type1 == type2:
                raise TypeError(
                    f'unsupported model type for *: {lhs} (additive) '
                    f'and {rhs} (additive)'
                )

            if 'conv' == type1 or 'conv' == type2:
                raise TypeError(
                    f'unsupported model type for *: {lhs} ({type1}) '
                    f'and {rhs} ({type2})'
                )

            self._type = 'add' if type1 == 'add' or type2 == 'add' else 'mul'
            self._op = jnp.multiply
            op_name = '{} * {}'
            op_latex = r'{{{}}} \times {{{}}}'

        else:
            raise NotImplementedError(f'op {op}')

        lhs_name = lhs._name
        lhs_latex = lhs._latex
        rhs_name = rhs._name
        rhs_latex = rhs._latex

        if op == '*':
            if isinstance(lhs, CompositeModel) and lhs._op_symbol == '+':
                lhs_name = f'({lhs_name})'
                lhs_latex = rf'\left({lhs_latex}\right)'

            if isinstance(rhs, CompositeModel) and rhs._op_symbol == '+':
                rhs_name = f'({rhs_name})'
                rhs_latex = rf'\left({rhs_latex}\right)'

        comps = list(lhs._comps)
        for c in rhs._comps:
            if c not in comps:
                comps.append(c)

        super().__init__(
            op_name.format(lhs_name, rhs_name),
            op_latex.format(lhs_latex, rhs_latex),
            comps,
        )

    @property
    def eval(self) -> ModelEval:
        op = self._op
        lhs = self._operands[0].eval
        rhs = self._operands[1].eval

        def fn(egrid: JAXArray, params: CompIDParamValMapping) -> JAXArray:
            """The model evaluation function"""
            return op(lhs(egrid, params), rhs(egrid, params))

        return jax.jit(fn)

    @property
    def _additive_comps(self) -> tuple[Model, ...]:
        if self.__additive_comps is not None:
            return self.__additive_comps

        if self.type != 'add':
            comps = ()
        else:
            op = self._op_symbol
            lhs = self._operands[0]
            rhs = self._operands[1]

            if op == '+':  # add + add
                comps = lhs._additive_comps + rhs._additive_comps
            else:
                if lhs.type == 'add':  # add * mul
                    lhs_comps = lhs._additive_comps
                    if len(lhs_comps) >= 2:
                        comps = tuple(c * rhs for c in lhs_comps)
                    else:
                        comps = (self,)
                else:  # mul * add
                    rhs_comps = rhs._additive_comps
                    if len(rhs_comps) >= 2:
                        comps = tuple(lhs * c for c in rhs_comps)
                    else:
                        comps = (self,)

        self.__additive_comps = comps

        return self.__additive_comps

    @property
    def type(self) -> Literal['add', 'mul']:
        return self._type


class ComponentMeta(ABCMeta):
    """Avoid cumbersome coding for subclass ``__init__``."""

    def __new__(metacls, name, bases, dct, **kwargs) -> ComponentMeta:
        config = dct.get('_config', None)

        # check config
        if isinstance(config, tuple):
            if not config:
                raise ValueError(f'empty parameter configurations for {name}')

            if not all(isinstance(cfg, ParamConfig) for cfg in config):
                raise ValueError(
                    f'each parameter config of {name} must be a ParamConfig'
                )

            for i, cfg in enumerate(config):
                dct[cfg.name] = property(
                    fget=_param_getter(cfg.name),
                    fset=_param_setter(cfg.name, i),
                    doc=f'{name} parameter :math:`{cfg.latex}`.',
                )

        # create the class
        cls = super().__new__(metacls, name, bases, dct)

        # check if all static methods are defined correctly
        if config is not None:
            for mname in cls._staticmethod:
                method = inspect.getattr_static(cls, mname)
                if not isinstance(method, staticmethod):
                    raise TypeError(f'{name}.{mname} must be a staticmethod')

        # define __init__ method of the newly created class if necessary
        if config is not None and '__init__' not in dct:
            # signature of __init__ method
            sig1 = [
                f"{i[0]}: 'Parameter' | float | None = None" for i in config
            ]
            sig1 += ['latex: str | None = None']

            params = '{' + ', '.join(f"'{i[0]}': {i[0]}" for i in config) + '}'

            # signature of super().__init__ method
            sig2 = [f'params={params}', 'latex=latex']

            if args := getattr(cls, '_args', None):
                sig1 = [f'{i}' for i in args] + sig1
                sig2 = [f'{i}={i}' for i in args] + sig2

            if kwargs := getattr(cls, '_kwargs', None):
                sig1 = sig1 + [f'{i}=None' for i in kwargs]
                sig2 = [f'{i}={i}' for i in kwargs] + sig2

            func_code = (
                f'def __init__(self, {", ".join(sig1)}):\n'
                f'    super(type(self), self).__init__({", ".join(sig2)})\n'
                f'__init__.__qualname__ = "{name}.__init__"'
            )
            exec(func_code, tmp := {})
            cls.__init__ = tmp['__init__']

        return cls

    def __init__(cls, name, bases, dct, **kwargs) -> None:
        super().__init__(name, bases, dct, **kwargs)

        # restore the signature of __init__ method
        # see https://stackoverflow.com/a/65385559
        signature = inspect.signature(cls.__init__)
        cls.__signature__ = signature.replace(
            parameters=tuple(signature.parameters.values())[1:],
            return_annotation=UniComponentModel,
        )

    def __call__(cls, *args, **kwargs) -> UniComponentModel | ConvolutionModel:
        # return Model/ConvolutionModel after Component initialization
        component = super().__call__(*args, **kwargs)

        if component.type != 'conv':
            return UniComponentModel(component)
        else:
            return ConvolutionModel(component)


def _param_getter(name: str) -> Callable[[Component], Parameter]:
    def getter(self: Component) -> Parameter:
        """Get parameter."""
        return getattr(self, f'__{name}')

    return getter


def _param_setter(name: str, idx: int) -> Callable[[Component, Any], None]:
    def setter(self: Component, param: Any) -> None:
        """Set parameter."""

        if param is not None and getattr(self, f'__{name}', None) is param:
            return

        cfg = self._config[idx]

        if param is None:
            # use default configuration
            param = UniformParameter(
                name=cfg.name,
                default=cfg.default,
                min=cfg.min,
                max=cfg.max,
                log=cfg.log,
                fixed=cfg.fixed,
                latex=cfg.latex,
            )

        elif isinstance(
            param, float | int | jnp.number | jnp.ndarray | np.ndarray
        ):
            # fixed to the given value
            if jnp.shape(param) != ():
                raise ValueError('scalar value is expected')

            param = UniformParameter(
                name=cfg.name,
                default=param,
                min=cfg.min,
                max=cfg.max,
                log=cfg.log,
                fixed=True,
                latex=cfg.latex,
            )

        elif isinstance(param, Parameter):
            # given parameter instance
            param = param
            # set the latex format as specified in model config
            param.latex = cfg.latex

        elif isinstance(param, Mapping):
            # given mapping

            if not {'default', 'min', 'max'}.issubset(param.keys()):
                raise ValueError(
                    f'{type(self).__name__}.{cfg.name} expected dict with keys'
                    f' "default", "min", "max", with optional "log" and'
                    f' "fixed", but got {param}'
                )

            param = UniformParameter(
                name=cfg.name,
                default=param['default'],
                min=param['min'],
                max=param['max'],
                log=param.get('log', False),
                fixed=param.get('fixed', False),
                latex=cfg.latex,
            )

        elif isinstance(param, Sequence):
            # given sequence
            if len(param) not in {1, 3, 4}:
                raise ValueError(
                    f'{type(self).__name__}.{cfg.name} expected sequence of '
                    'length 1, 3, or 4: '
                    '[default (float)], '
                    '[default (float), min (float), max (float)], or '
                    '[default (float), min (float), max (float), log (bool)], '
                    f'but got {param}'
                )

            if len(param) == 1:
                (default,) = param
                min_ = cfg.min
                max_ = cfg.max
                log = cfg.log
            elif len(param) == 3:
                default, min_, max_ = param
                log = cfg.log
            else:
                default, min_, max_, log = param

            param = UniformParameter(
                name=cfg.name,
                default=default,
                min=min_,
                max=max_,
                log=log,
                fixed=cfg.fixed,
                latex=cfg.latex,
            )

        else:
            # other input types are not supported yet
            raise TypeError(
                f'{type(self).__name__}.{cfg.name} got unsupported value '
                f'{param} ({type(param).__name__})'
            )

        prev_param: Parameter | None = getattr(self, f'__{name}', None)
        if prev_param is not None:
            prev_param._tracker.remove(self._id, name)

        setattr(self, f'__{name}', param)

        param._tracker.append(self._id, name)

    return setter


class ParamConfig(NamedTuple):
    """Configuration of a uniform parameter for spectral component."""

    name: str
    """Plain name of the parameter."""

    latex: str
    r""":math:`\LaTeX` of the parameter."""

    unit: str
    """Physical unit."""

    default: float
    """Default value of the parameter."""

    min: float
    """Minimum value of the parameter."""

    max: float
    """Maximum value of the parameter."""

    log: bool = False
    """Whether the parameter is parameterized in a logarithmic scale."""

    fixed: bool = False
    """Whether the parameter is fixed."""


class Component(ABC, metaclass=ComponentMeta):
    """Base class to define a spectral component."""

    _config: tuple[ParamConfig, ...]
    _id: CompID
    _args: tuple[str, ...] = ()  # extra args passed to subclass __init__
    _kwargs: tuple[str, ...] = ()  # extra kwargs passed to subclass __init__
    _staticmethod: tuple[str, ...] = ()  # methods need to be static
    __initialized: bool = False

    def __init__(self, params: dict, latex: str | None):
        self._id = hex(id(self))[2:]

        if latex is None:
            latex = rf'\mathrm{{{self.__class__.__name__}}}'
        self.latex = latex

        self._name = self.__class__.__name__

        # parse parameters from params, which is a dict of parameters
        for cfg in self._config:
            setattr(self, cfg.name, params[cfg.name])

        self._param_names = tuple(cfg.name for cfg in self._config)

        self.__initialized = True

    @property
    @abstractmethod
    def eval(self) -> CompEval:
        """Get side-effect free component evaluation function."""
        pass

    @property
    def name(self) -> str:
        """Component name."""
        return self._name

    @property
    def latex(self) -> str:
        r""":math:`\LaTeX` format of the component."""
        return self._latex

    @latex.setter
    def latex(self, latex: str):
        self._latex = str(latex)

    @property
    @abstractmethod
    def type(self) -> Literal['add', 'mul', 'conv']:
        """Component type."""
        pass

    @property
    def param_names(self) -> tuple[str, ...]:
        """Component's parameter names."""
        return self._param_names

    @property
    @abstractmethod
    def _config(self) -> tuple[ParamConfig, ...]:
        """Default configuration of parameters."""
        pass

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        # TODO: show the component name, LaTeX, type and parameters
        return self.name

    def __getitem__(self, key: str) -> Parameter:
        if key not in self.param_names:
            raise KeyError(key)

        return getattr(self, key)

    def __setitem__(self, key: str, value: Any):
        if key not in self.param_names:
            raise KeyError(key)

        setattr(self, key, value)

    def __delattr__(self, key: str):
        if self.__initialized and hasattr(self, key):
            raise AttributeError("can't delete attribute")

        super().__delattr__(key)

    def __setattr__(self, key: str, value: Any):
        if self.__initialized and not hasattr(self, key):
            raise AttributeError("can't set attribute")

        super().__setattr__(key, value)


class AdditiveComponent(Component):
    """Prototype class to define an additive component."""

    @property
    def type(self) -> Literal['add']:
        """Component type is additive."""
        return 'add'


class MultiplicativeComponent(Component):
    """Prototype class to define a multiplicative component."""

    @property
    def type(self) -> Literal['mul']:
        """Component type is multiplicative."""
        return 'mul'


class AnalyticalIntegral(Component):
    """Prototype component to calculate model integral analytically."""

    _integral_jit: CompEval | None = None
    _staticmethod = ('integral',)

    def __init__(self, params: dict, latex: str | None):
        super().__init__(params, latex)

    @property
    def eval(self) -> CompEval:
        if self._integral_jit is None:
            self._integral_jit = jax.jit(self.integral)

        return self._integral_jit

    @staticmethod
    @abstractmethod
    def integral(*args, **kwargs) -> JAXArray:
        """Calculate the model value over grid."""
        pass


class NumericalIntegral(Component):
    """Prototype component to calculate model integral numerically."""

    _kwargs = ('method',)
    _continuum_jit: CompEval | None = None
    _staticmethod = ('continuum',)

    def __init__(
        self,
        params: dict,
        latex: str | None,
        method: Literal['trapz', 'simpson'] | None = None,
    ):
        self.method = 'trapz' if method is None else method
        super().__init__(params, latex)

    @property
    def eval(self) -> CompEval:
        if self._continuum_jit is None:
            # _continuum is assumed to be a pure function, independent of self
            self._continuum_jit = jax.jit(self.continuum)

        return self._make_integral(self._continuum_jit)

    def _make_integral(self, continuum: CompEval):
        mtype = self.type

        if self.method == 'trapz':

            def fn(egrid: JAXArray, params: NameValMapping) -> JAXArray:
                """Numerical integration using trapezoidal rule."""
                if mtype == 'add':
                    factor = 0.5 * (egrid[1:] - egrid[:-1])
                else:
                    factor = 0.5
                f_grid = continuum(egrid, params)
                return factor * (f_grid[:-1] + f_grid[1:])

        elif self.method == 'simpson':

            def fn(egrid: JAXArray, params: NameValMapping) -> JAXArray:
                """Numerical integration using Simpson's 1/3 rule."""
                if mtype == 'add':
                    factor = (egrid[1:] - egrid[:-1]) / 6.0
                else:
                    factor = 1.0 / 6.0
                e_mid = 0.5 * (egrid[:-1] + egrid[1:])
                f_grid = continuum(egrid, params)
                f_mid = continuum(e_mid, params)
                return factor * (f_grid[:-1] + 4.0 * f_mid + f_grid[1:])

        else:
            raise NotImplementedError(f"integration method '{self.method}'")

        return jax.jit(fn)

    @staticmethod
    @abstractmethod
    def continuum(*args, **kwargs) -> JAXArray:
        """Calculate the model value at the energy grid."""
        pass

    @property
    def method(self) -> Literal['trapz', 'simpson']:
        """Numerical integration method."""
        return self._method

    @method.setter
    def method(self, method: Literal['trapz', 'simpson']):
        if method not in ('trapz', 'simpson'):
            raise ValueError(
                f"available integration methods are 'trapz' and 'simpson', "
                f"but got '{method}'"
            )

        self._method = method


class AnaIntAdditive(AnalyticalIntegral, AdditiveComponent):
    """Prototype additive component with integral expression defined."""

    @staticmethod
    @abstractmethod
    def integral(*args, **kwargs) -> JAXArray:
        """Calculate the photon flux integrated over the energy grid.

        Parameters
        ----------
        egrid : ndarray
            Photon energy grid in units of keV.
        params : dict
            Parameter dict for the model.

        Returns
        -------
        jax.Array
            The photon flux integrated over `egrid`, in units of ph cm⁻² s⁻¹.
        """
        pass


class NumIntAdditive(NumericalIntegral, AdditiveComponent):
    """Prototype additive component with continuum expression defined."""

    @staticmethod
    @abstractmethod
    def continuum(*args, **kwargs) -> JAXArray:
        """Calculate the photon flux at the energy grid.

        Parameters
        ----------
        egrid : ndarray
            Photon energy grid in units of keV.
        params : dict
            Parameter dict for the model.

        Returns
        -------
        jax.Array
            The photon flux at `egrid`, in units of ph cm⁻² s⁻¹ keV⁻¹.
        """
        pass


class AnaIntMultiplicative(AnalyticalIntegral, MultiplicativeComponent):
    """Prototype multiplicative component with integral expression defined."""

    @staticmethod
    @abstractmethod
    def integral(*args, **kwargs) -> JAXArray:
        """Calculate the average value of the model between the grid.

        Parameters
        ----------
        egrid : ndarray
            Photon energy grid in units of keV.
        params : dict
            Parameter dict for the model.

        Returns
        -------
        jax.Array
            The model value, dimensionless.
        """
        pass


class NumIntMultiplicative(NumericalIntegral, MultiplicativeComponent):
    """Prototype multiplicative component with continuum expression defined."""

    @staticmethod
    @abstractmethod
    def continuum(*args, **kwargs) -> JAXArray:
        """Calculate the model value at the energy grid.

        Parameters
        ----------
        egrid : ndarray
            Photon energy grid in units of keV.
        params : dict
            Parameter dict for the model.

        Returns
        -------
        jax.Array
            The model value at `egrid`, dimensionless.
        """
        pass


class ConvolutionModel(Model):
    """Model defined by a convolution component."""

    def __init__(self, component: ConvolutionComponent):
        self._component = component
        super().__init__(component._id, component._id, [component])

    def __call__(self, model: Model) -> ConvolvedModel:
        if model.type not in self._component._supported:
            accepted = [f"'{i}'" for i in self._component._supported]
            raise TypeError(
                f'{self.name} convolution model supports models with type: '
                f"{', '.join(accepted)}; got '{model.type}' type model {model}"
            )

        return ConvolvedModel(self._component, model)

    @property
    def eval(self):
        # this is not supposed to be called
        raise RuntimeError('convolution model has no eval defined')

    @property
    def type(self) -> Literal['conv']:
        """Model type is convolution."""
        return 'conv'


class ConvolvedModel(Model):
    """Model created by convolution."""

    _op: ConvolutionComponent
    _model: Model

    def __init__(self, op: ConvolutionComponent, model: Model):
        self._op = op
        self._model = model
        name = f'{op._id}({model._name})'
        latex = rf'{{{op._id}}}\left({model._latex}\right)'

        comps = list(model._comps)
        if op not in comps:
            comps = comps + [op]

        self.__additive_comps = None

        super().__init__(name, latex, comps)

    @property
    def _additive_comps(self) -> tuple[Model, ...]:
        if self.__additive_comps is not None:
            return self.__additive_comps

        if self.type != 'add':
            comps = ()
        else:
            op = self._op
            model_comps = self._model._additive_comps

            if len(model_comps) >= 2:
                comps = tuple(ConvolvedModel(op, m) for m in model_comps)
            else:
                comps = (self,)

        self.__additive_comps = comps

        return self.__additive_comps

    @property
    def eval(self) -> ModelEval:
        comp_id = self._op._id
        _fn = self._op.eval
        _model_fn = self._model.eval

        def fn(egrid: JAXArray, params: CompIDParamValMapping) -> JAXArray:
            """The convolved model evaluation function."""
            return _fn(egrid, params[comp_id], lambda e: _model_fn(e, params))

        return jax.jit(fn)

    @property
    def type(self) -> Literal['add', 'mul']:
        return self._model.type


class ConvolutionComponent(Component):
    """Prototype class to define a convolution component."""

    _supported: frozenset = frozenset({'add', 'mul'})
    _convolve_jit: ConvolveEval | None = None
    _staticmethod = ('convolve',)

    @property
    def type(self) -> Literal['conv']:
        return 'conv'

    @property
    def eval(self) -> ConvolveEval:
        if self._convolve_jit is None:
            self._convolve_jit = jax.jit(self.convolve, static_argnums=2)

        return self._convolve_jit

    @staticmethod
    @abstractmethod
    def convolve(*args, **kwargs) -> JAXArray:
        """Convolve a model function.

        Parameters
        ----------
        egrid : ndarray
            Photon energy grid in units of keV.
        params : dict
            Parameter dict for the convolution model.
        model_fn : callable
            The model function to be convolved, which takes the energy grid as
            input and returns the model value over the grid.

        Returns
        -------
        jax.Array
            The convolved model value over `egrid`.
        """
        pass


class PyComponent(Component):
    """Prototype component with pure Python expression defined."""

    _kwargs: tuple[str, ...] = ('grad_method',)

    def __init__(
        self,
        params: dict,
        latex: str | None,
        grad_method: Literal['central', 'forward'] | None,
    ):
        self.grad_method = grad_method

        super().__init__(params, latex)

    @property
    def grad_method(self) -> Literal['central', 'forward']:
        """Numerical differentiation method."""
        return self._grad_method

    @grad_method.setter
    def grad_method(self, value: Literal['central', 'forward'] | None):
        if value is None:
            value: Literal['central'] = 'central'

        if value not in {'central', 'forward'}:
            raise ValueError(
                f"supported methods are 'central' and 'forward', but got "
                f"'{value}'"
            )
        self._grad_method = value


class PyAnaInt(PyComponent, AnalyticalIntegral):
    """Prototype component with python integral expression defined."""

    @property
    def eval(self) -> CompEval:
        if self._integral_jit is None:
            integral_fn = self.integral

            def eval_integral(egrid, params):
                egrid = np.asarray(egrid)
                params = {k: np.asarray(v) for k, v in params.items()}
                return integral_fn(egrid, params)

            def integral(egrid: JAXArray, params: NameValMapping) -> JAXArray:
                shape_dtype = jax.ShapeDtypeStruct(
                    (egrid.size - 1,), egrid.dtype
                )
                return jax.pure_callback(
                    eval_integral,
                    shape_dtype,
                    egrid,
                    params,
                    vmap_method='sequential',
                )

            self._integral_jit = jax.jit(
                define_fdjvp(jax.jit(integral), self.grad_method)
            )

        return self._integral_jit


class PyNumInt(PyComponent, NumericalIntegral):
    """Prototype component with python continuum expression defined."""

    _kwargs = ('method', 'grad_method')

    def __init__(
        self,
        params: dict,
        latex: str | None,
        method: Literal['trapz', 'simpson'] | None,
        grad_method: Literal['central', 'forward'] | None,
    ):
        super().__init__(params, latex, grad_method)
        self.method = 'trapz' if method is None else method

    @property
    def eval(self) -> CompEval:
        if self._continuum_jit is None:
            # continuum is assumed to be a pure function, independent of self
            continuum_fn = self.continuum

            def eval_continuum(egrid, params):
                egrid = np.asarray(egrid)
                params = {k: np.asarray(v) for k, v in params.items()}
                return continuum_fn(egrid, params)

            def continuum(egrid: JAXArray, params: NameValMapping) -> JAXArray:
                shape_dtype = jax.ShapeDtypeStruct(egrid.shape, egrid.dtype)
                return jax.pure_callback(
                    eval_continuum,
                    shape_dtype,
                    egrid,
                    params,
                    vmap_method='sequential',
                )

            self._continuum_jit = jax.jit(
                define_fdjvp(jax.jit(continuum), self.grad_method)
            )

        return self._make_integral(self._continuum_jit)


def get_mtype_str(model: Model | CompiledModel) -> str:
    """Get the model type string."""
    match model.type:
        case 'add':
            return 'Additive'
        case 'mul':
            return 'Multiplicative'
        case 'conv':
            return 'Convolution'
        case _:
            raise ValueError(f"unknown model type '{model.type}'")


def get_model_info(
    comps: Sequence[Component],
    cid_to_name: CompIDStrMapping,
    cid_to_latex: CompIDStrMapping,
) -> ModelInfo:
    """Get the model information.

    Parameters
    ----------
    comps : sequence of Component
        The sequence of components.
    cid_to_name : mapping
        The mapping of component id to component name.
    cid_to_latex : mapping
        The mapping of component id to component LaTeX format.

    Returns
    -------
    ModelInfo
        The model parameter information dict.
    """
    # get all parameter information
    params_info: dict[ParamID, ParamInfo] = reduce(
        lambda x, y: x | y,
        [comp[name]._info for comp in comps for name in comp.param_names],
    )

    # mapping from pid to assigned component id and parameter name
    comp_param: dict[ParamID, tuple[CompID, CompParamName] | None] = {
        pid: tmp
        for pid, info in params_info.items()
        if (tmp := info.tracker.get_comp_param(cid_to_name.keys()))
    }

    # names of non-composite params with no component assigned (aux params)
    aux_params = {
        pid: info
        for pid, info in params_info.items()
        if (pid not in comp_param) and (not info.composite)
    }

    aux_names = [i.name for i in aux_params.values()]
    aux_names = build_namespace(aux_names, prime=True)['namespace']

    # name mapping from aux params id to name
    name_mapping: ParamIDStrMapping = dict(
        zip(aux_params.keys(), aux_names, strict=True)
    )

    # record name mapping of params directly assigned to a component
    name_mapping |= {
        pid: f'{cid_to_name[cid]}.{pname}'
        for pid, (cid, pname) in comp_param.items()
    }

    # record the component LaTeX format of parameters
    comp_latex_mapping = {
        pid: cid_to_latex[cid] for pid, (cid, _) in comp_param.items()
    }
    comp_latex_mapping |= dict.fromkeys(aux_params, '')

    # record the LaTeX format of aux parameters
    aux_latex = [params_info[pid].latex for pid in aux_params]
    aux_latex = build_namespace(aux_latex, latex=True, prime=True)['namespace']
    latex_mapping = dict(zip(aux_params.keys(), aux_latex, strict=True))

    # record the LaTeX format of component parameters from _config
    # latex_mapping |= {
    #     comp[name]._id: comp._config[i].latex
    #     for comp in comps
    #     for (i, name) in enumerate(comp.param_names)
    # }

    # record the LaTeX format of component parameters
    latex_mapping |= {
        comp[name]._id: params_info[comp[name]._id].latex
        for comp in comps
        for (i, name) in enumerate(comp.param_names)
    }

    # TODO: record aux params unit & unit consistency check
    # record the unit of component parameters from _config
    unit_mapping = {
        comp[name]._id: comp._config[i].unit
        for comp in comps
        for (i, name) in enumerate(comp.param_names)
    }
    unit_mapping |= {pid: '' for pid, info in aux_params.items()}

    # record whether the parameter is logarithmic
    log = {pid: params_info[pid].log for pid in name_mapping}

    # record the value of fixed parameters
    fixed = {
        pid: info.default
        for pid, info in params_info.items()
        if (pid in name_mapping) and info.fixed and (not info.integrate)
    }

    # integral operator
    integrate = {
        pid: info.integrate
        for pid, info in params_info.items()
        if info.integrate and not info.composite
    }

    # record the sample distribution of free parameters
    sample: dict[ParamID, Distribution] = {
        pid: info.dist for pid, info in params_info.items() if info.dist
    }

    # record the default value of free parameters
    default: dict[ParamID, JAXFloat] = {
        pid: params_info[pid].default for pid in sample
    }

    # ========== generate component parameter value getter function ===========
    def factory1(
        id_to_value: Callable[[ParamIDValMapping], JAXFloat],
    ) -> Callable[[ParamIDValMapping], JAXFloat]:
        """Fill in the default parameter values for fn."""

        def fn(value_mapping: ParamIDValMapping) -> JAXFloat:
            """id_to_value with default values."""
            return id_to_value(fixed_values | value_mapping)

        return fn

    def factory2(
        pname_to_pid: dict[CompParamName, ParamID],
    ) -> Callable[[ParamIDValMapping], NameValMapping]:
        """Generate component parameter value getter."""

        def fn(value_mapping: ParamIDValMapping) -> NameValMapping:
            """Component parameter value getter."""
            return {
                pname: pid_to_value[pid](value_mapping)
                for pname, pid in pname_to_pid.items()
            }

        return fn

    fixed_values = fixed
    pid_to_value = {
        pid: factory1(params_info[pid].id_to_value) for pid in comp_param
    }
    cid_to_params = {
        comp._id: factory2({name: comp[name]._id for name in comp.param_names})
        for comp in comps
    }
    # ========== generate component parameter value getter function ===========

    # record non-interval params which has a component assigned
    deterministic = {}

    # record the setup of component parameters
    setup: dict[CompParamName, (ParamID, ParamSetup)] = {}

    # model_info is list of (No., Component, Parameter, Value, Bound, Prior)
    model_info = []
    sample_order = []
    n = 0

    for comp in comps:  # iterate through components
        cid = comp._id
        cname = cid_to_name[cid]
        for pname in comp.param_names:  # iterate through parameters of comp
            model_pname = f'{cname}.{pname}'
            pid = comp[pname]._id
            info = params_info[pid]

            idx = ''
            bound = info.bound
            prior = info.prior

            if (cid, pname) != comp_param[pid]:  # param is linked to another
                idx = ''
                value = name_mapping[pid]
                setup[model_pname] = (value, ParamSetup.Forwarded)

            elif info.composite:  # param is composite
                mapping = dict(name_mapping)
                del mapping[pid]
                value = info.name(mapping)

                if info.integrate:  # param is composite interval
                    setup[model_pname] = (model_pname, ParamSetup.Integrated)
                elif info.fixed:  # param is composite, but fixed
                    setup[model_pname] = (model_pname, ParamSetup.Fixed)
                else:  # param is composite but free to vary, add it to determ
                    idx = '*'
                    setup[model_pname] = (model_pname, ParamSetup.Composite)
                    deterministic[pid] = pid_to_value[pid]

            elif info.integrate:  # param is interval
                value = f'[{info.default[0]:.4g}, {info.default[1]:.4g}]'
                setup[model_pname] = (model_pname, ParamSetup.Integrated)

            else:  # a single param
                value = f'{info.default:.4g}'

                if info.fixed:
                    setup[model_pname] = (model_pname, ParamSetup.Fixed)
                else:
                    setup[model_pname] = (model_pname, ParamSetup.Free)
                    n += 1
                    idx = str(n)
                    sample_order.append(pid)

            model_info.append((idx, cname, pname, value, bound, prior))

    for pid, info in aux_params.items():  # iterate through aux params
        idx = ''
        bound = info.bound
        prior = info.prior

        if info.integrate:  # aux param is interval
            value = f'[{info.default[0]:.4g}, {info.default[1]:.4g}]'
        else:  # aux param is not interval, record its default value
            value = f'{info.default:.4g}'

        if not info.fixed:  # aux param is free to vary
            n += 1
            idx = str(n)
            sample_order.append(pid)

        model_info.append((idx, '', name_mapping[pid], value, bound, prior))

    if set(sample_order) != set(sample):
        raise RuntimeError('sample_order and sample are not consistent')

    sample = {pid: sample[pid] for pid in sample_order}

    return ModelInfo(
        info=tuple(model_info),
        name=name_mapping,
        latex=latex_mapping,
        unit=unit_mapping,
        sample=sample,
        default=default,
        deterministic=deterministic,
        fixed=fixed,
        log=log,
        pid_to_comp_latex=comp_latex_mapping,
        cid_to_params=cid_to_params,
        cid_to_name=cid_to_name,
        cid_to_latex=cid_to_latex,
        integrate=integrate,
        setup=setup,
    )


class ModelInfo(NamedTuple):
    """Model information."""

    info: tuple[tuple[str, str, str, str, str, str], ...]
    """The model parameter information.

    Each row contains (No., Component, Parameter, Value, Bound, Prior).
    """

    name: ParamIDStrMapping
    """The mapping of parameter id to parameter name."""

    latex: ParamIDStrMapping
    r"""The mapping of component parameters name to :math:`\LaTeX` format."""

    unit: ParamIDStrMapping
    """The mapping of component parameters name to physical unit."""

    sample: dict[ParamID, Distribution]
    """The mapping of free parameter id to numpyro Distribution."""

    default: ParamIDValMapping
    """The mapping of free parameter id to default value."""

    deterministic: dict[ParamID, Callable[[ParamIDValMapping], JAXFloat]]
    """The mapping of deterministic parameters id to value getter."""

    fixed: ParamIDValMapping
    """The mapping of fixed parameter id to fixed parameter value."""

    log: dict[ParamID, bool]
    """The mapping of parameter id to logarithmic flag."""

    pid_to_comp_latex: dict[ParamID, str]
    """The mapping of parameter id to component LaTeX format."""

    cid_to_name: dict[CompID, CompName]
    """The mapping of component id to component name."""

    cid_to_latex: dict[CompID, str]
    """The mapping of component id to component LaTeX format."""

    cid_to_params: dict[CompID, Callable[[ParamIDValMapping], NameValMapping]]
    """The mapping of component id to parameter value getter function."""

    integrate: dict[ParamID, IntegralFactory]
    """The mapping of interval parameter id to integral operator."""

    setup: dict[CompParamName, tuple[ParamID, ParamSetup]]
    """The mapping of component parameter setup."""


class ParamSetup(Enum):
    """Model parameter setup."""

    Free = 0
    """Parameter is free to vary."""

    Composite = 1
    """Parameter is composed of other free parameters."""

    Forwarded = 2
    """Parameter is directly forwarded to another model parameter."""

    Fixed = 3
    """Parameter is fixed to a value."""

    Integrated = 4
    """Parameter is integrated out."""



"""Miscellaneous helper functions."""

from __future__ import annotations

import math
import re
from functools import reduce
from threading import Lock
from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp
from astropy.units import Unit
from jax import lax
from jax.custom_derivatives import SymbolicZero
from jax.experimental import io_callback
from jax.flatten_util import ravel_pytree
from prettytable import PrettyTable
from tqdm.auto import tqdm

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from typing import Literal, TypeVar

    from numpy import ndarray as NDArray

    from elisa.util.typing import CompEval

    T = TypeVar('T')

UNICODE_SUBSCRIPT = dict(
    zip(
        'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+-/=()',
        'ᴀʙᴄᴅᴇғɢʜɪᴊᴋʟᴍɴᴏᴘǫʀsᴛᴜᴠᴡxʏᴢₐᵦ𝒸𝒹ₑ𝒻𝓰ₕᵢⱼₖₗₘₙₒₚᵩᵣₛₜᵤᵥ𝓌ₓᵧ𝓏₀₁₂₃₄₅₆₇₈₉₊₋⸝₌₍₎',
        strict=True,
    )
)
UNICODE_SUPERSCRIPT = dict(
    zip(
        'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+-/=()',
        'ᴬᴮᶜᴰᴱᶠᴳᴴᴵᴶᴷᴸᴹᴺᴼᴾᵠᴿˢᵀᵁⱽᵂᕽʸᶻᵃᵇᶜᵈᵉᶠᵍʰⁱʲᵏˡᵐⁿᵒᵖᵠʳˢᵗᵘᵛʷˣʸᶻ⁰¹²³⁴⁵⁶⁷⁸⁹⁺⁻ᐟ⁼⁽⁾',
        strict=True,
    )
)
UNICODE_SUFFIX = False
PLAIN_SUBSCRIPT_TEMPLATE = '_%s'
PLAIN_SUPERSCRIPT_TEMPLATE = '[%s]'
LATEX_SUBSCRIPT_TEMPLATE = '_{%s}'
LATEX_SUPERSCRIPT_TEMPLATE = '^{%s}'


def set_suffix_config(
    unicode: bool = False,
    plain_subscript_template: str = '_%s',
    plain_superscript_template: str = '[%s]',
    latex_subscript_template: str = '_{%s}',
    latex_superscript_template: str = '^{%s}',
) -> None:
    """Set suffix configuration.

    Parameters
    ----------
    unicode : bool, optional
        If True, use unicode suffix. The default is False.
    plain_subscript_template : str, optional
        The template for plain subscript. The default is ``'_%s'``.
    plain_superscript_template : str, optional
        The template for plain superscript. The default is ``'[%s]'``.
    latex_subscript_template : str, optional
        The template for LaTeX subscript. The default is ``'_{%s}'``.
    latex_superscript_template : str, optional
        The template for LaTeX superscript. The default is ``'^{%s}'``.
    """
    # test if templates are valid
    try:
        _ = plain_subscript_template % 'test'
        _ = plain_superscript_template % 'test'
        _ = latex_subscript_template % 'test'
        _ = latex_superscript_template % 'test'
    except TypeError as e:
        raise ValueError(
            'subscript and superscript templates must be valid '
            'Python string formatting templates'
        ) from e

    global UNICODE_SUFFIX
    global PLAIN_SUBSCRIPT_TEMPLATE
    global PLAIN_SUPERSCRIPT_TEMPLATE
    global LATEX_SUBSCRIPT_TEMPLATE
    global LATEX_SUPERSCRIPT_TEMPLATE
    UNICODE_SUFFIX = bool(unicode)
    PLAIN_SUBSCRIPT_TEMPLATE = plain_subscript_template
    PLAIN_SUPERSCRIPT_TEMPLATE = plain_superscript_template
    LATEX_SUBSCRIPT_TEMPLATE = latex_subscript_template
    LATEX_SUPERSCRIPT_TEMPLATE = latex_superscript_template


def add_suffix(
    strings: str | Sequence[str],
    suffix: str | Sequence[str],
    subscript: bool,
    unicode: bool | None = None,
    latex: bool = False,
    mathrm: bool = False,
) -> str | list[str]:
    """Add suffix to a sequence of strings.

    Parameters
    ----------
    strings : sequence of str
        The sequence of strings.
    suffix : sequence of str
        The sequence of suffixes. The suffix format can be set by
        :py:func:`elisa.util.misc.set_suffix_config`.
    subscript : bool, optional
        If True, add suffix as subscript, otherwise superscript.
        The default is True.
    latex : bool, optional
        If True, add suffix following LaTeX format. The default is False.
    unicode : bool, optional
        If True, add suffix with Unicode string. Defaults to
        ``elisa.util.misc.UNICODE_SUFFIX``, which can be set by
        :py:func:`elisa.util.misc.set_suffix_config`.
    mathrm : bool, optional
        If True, add suffix in mathrm when latex is True. The default is False.

    Returns
    -------
    str or list of str
        The strings with suffix added.
    """
    if unicode is None:
        unicode = UNICODE_SUFFIX

    return_list = False

    if isinstance(strings, str):
        strings = [strings]
    else:
        strings = list(strings)
        return_list = True

    if isinstance(suffix, str):
        suffix = [suffix]
    else:
        suffix = list(suffix)
        return_list = True

    if len(strings) != len(suffix):
        raise ValueError('length of `strings` and `suffix` must be the same')

    def to_unicode(string: str):
        """Replace suffix with unicode."""
        if subscript:
            return ''.join(f'{UNICODE_SUBSCRIPT.get(i, i)}' for i in string)
        else:
            return ''.join(f'{UNICODE_SUPERSCRIPT.get(i, i)}' for i in string)

    if latex:
        if subscript:
            template = LATEX_SUBSCRIPT_TEMPLATE
        else:
            template = LATEX_SUPERSCRIPT_TEMPLATE
        if mathrm:
            suffix = [r'\mathrm{' + i + '}' if i else '' for i in suffix]
        else:
            suffix = ['{' + i + '}' if i else '' for i in suffix]
        strings = ['{' + i + '}' for i in strings]
    elif unicode:
        template = '%s'
        suffix = [to_unicode(i) for i in suffix]
    else:
        if subscript:
            template = PLAIN_SUBSCRIPT_TEMPLATE
        else:
            template = PLAIN_SUPERSCRIPT_TEMPLATE

    strings = [
        i + template % j if j else i
        for i, j in zip(strings, suffix, strict=True)
    ]

    if return_list:
        return strings
    else:
        return strings[0]


def build_namespace(
    names: Sequence[str],
    latex: bool = False,
    prime: bool = False,
) -> dict[str, list[str | int]]:
    """Build a namespace from a sequence of names.

    Parameters
    ----------
    names : sequence of str
        A sequence of names.
    latex : bool, optional
        If True, `names` are assumed to be LaTeX strings. The default is False.
    prime : bool, optional
        If True, primes are used as suffix for duplicate names, otherwise
        a number is used. The default is False.

    Returns
    -------
    namespace: dict
        A dict of non-duplicate names and suffixes in original name order.
    """
    namespace = []
    suffixes_n = []
    counter = {}

    for name in names:
        if name not in namespace:
            counter[name] = 1
            namespace.append(name)
        else:
            counter[name] += 1

        suffixes_n.append(counter[name])

    if prime:
        suffixes = [i - 1 for i in suffixes_n]
        suffixes = ["'" * n for n in suffixes]
    else:
        template = '_{%d}' if latex else '_%d'
        suffixes = [template % n if n > 1 else '' for n in suffixes_n]

    return {
        'namespace': list(map(''.join, zip(names, suffixes, strict=True))),
        'suffix_num': [str(n) if n > 1 else '' for n in suffixes_n],
    }


def define_fdjvp(
    fn: CompEval,
    method: Literal['central', 'forward'] = 'central',
) -> CompEval:
    """Define JVP using finite differences."""
    if method not in {'central', 'forward'}:
        raise ValueError(
            f"supported methods are 'central' and 'forward', but got "
            f"'{method}'"
        )

    def fdjvp(primals, tangents):
        egrid, params = primals
        egrid_tangent, params_tangent = tangents

        if not isinstance(egrid_tangent, SymbolicZero):
            raise NotImplementedError('JVP for energy grid is not implemented')

        primals_out = fn(egrid, params)

        tvals, _ = jax.tree.flatten(params_tangent)
        if any(jnp.shape(v) != () for v in tvals):
            raise NotImplementedError(
                'JVP for non-scalar parameter is not implemented'
            )

        non_zero_tangents = [not isinstance(v, SymbolicZero) for v in tvals]
        idx = [i for i, v in enumerate(non_zero_tangents) if v]
        idx_arr = jnp.array(idx)
        nbatch = sum(non_zero_tangents)
        nparam = len(tvals)
        params_ravel, revert = ravel_pytree(params)
        free_params_values = params_ravel[idx_arr]
        free_params_abs = jnp.where(
            jnp.equal(free_params_values, 0.0),
            jnp.ones_like(free_params_values),
            jnp.abs(free_params_values),
        )
        free_params_abs = jnp.expand_dims(free_params_abs, axis=-1)
        row_idx = jnp.arange(nbatch)
        perturb_idx = jnp.zeros((nbatch, nparam)).at[row_idx, idx_arr].set(1.0)
        params_batch = jnp.full((nbatch, nparam), params_ravel)

        eps = jnp.finfo(egrid.dtype).eps
        f_vmap = jax.vmap(fn, in_axes=(None, 0), out_axes=0)
        revert = jax.vmap(revert, in_axes=0, out_axes=0)

        # See Numerical Recipes Chapter 5.7
        if method == 'central':
            perturb = free_params_abs * eps ** (1.0 / 3.0)
            params_pos_perturb = revert(params_batch + perturb_idx * perturb)
            out_pos_perturb = f_vmap(egrid, params_pos_perturb)
            params_neg_perturb = revert(params_batch - perturb_idx * perturb)
            out_neg_perturb = f_vmap(egrid, params_neg_perturb)
            d_out = (out_pos_perturb - out_neg_perturb) / (2.0 * perturb)
        else:
            perturb = free_params_abs * jnp.sqrt(eps)
            params_perturb = revert(params_batch + perturb_idx * perturb)
            out_perturb = f_vmap(egrid, params_perturb)
            d_out = (out_perturb - primals_out) / perturb

        free_params_tangent = jnp.array([tvals[i] for i in idx])
        tangents_out = free_params_tangent @ d_out
        return primals_out, tangents_out

    fn = jax.custom_jvp(fn)
    fn.defjvp(fdjvp, symbolic_zeros=True)

    return fn


def get_unit_latex(unit: str, throw: bool = True) -> str:
    """Get latex string of a unit.

    Parameters
    ----------
    unit : str
        The unit string.
    throw : bool, optional
        If True, raise ValueError if the unit is invalid. The default is True.

    Returns
    -------
    str
        The latex string of the unit.
    """
    ustr = str(unit)
    if ustr:
        try:
            unit = Unit(ustr)
            max_index = len(ustr)
            pattern = r'(?:[^a-zA-Z]*){}(?:[^a-zA-Z]*)'
            index = [
                min(
                    (
                        r.start(0)
                        if (r := re.search(pattern.format(s), ustr))
                        is not None
                        else max_index
                    )
                    for s in [b.name] + b.aliases
                )
                for b in unit.bases
            ]
            index = sorted(range(len(index)), key=index.__getitem__)
            bases = [unit.bases[i].name for i in index]
            powers = [unit.powers[i] for i in index]
            ustr = r'\ '.join(
                b + (f'^{{{p}}}' if p != 1 else '')
                for b, p in zip(bases, powers, strict=True)
            )
            scale = Unit(unit.scale).to_string('latex_inline')[9:-2]
            if scale != '':
                scale = scale.replace(r'1 \times ', '')
                scale += r'\ '
            ustr = rf'$\mathrm{{{scale}{ustr}}}$'
        except ValueError as ve:
            if throw:
                raise ve
            ustr = ''

    return ustr


def make_pretty_table(fields: Sequence[str], rows: Sequence) -> PrettyTable:
    """Make a :class:`prettytable.PrettyTable`.

    Parameters
    ----------
    fields : sequence of str
        The names of fields.
    rows : sequence
        The sequence of data corresponding to the `fields`.

    Returns
    -------
    table : PrettyTable
        The pretty table.
    """
    table = PrettyTable(
        fields,
        align='c',
        hrules=1,  # 1 for all, 0 for frame
        vrules=1,
        padding_width=1,
        vertical_char='│',
        horizontal_char='─',
        junction_char='┼',
        top_junction_char='┬',
        bottom_junction_char='┴',
        right_junction_char='┤',
        left_junction_char='├',
        top_right_junction_char='╮',
        top_left_junction_char='╭',
        bottom_right_junction_char='╯',
        bottom_left_junction_char='╰',
    )
    table.add_rows(rows)
    return table


def replace_string(value: T, mapping: dict[str, str]) -> T:
    """Replace all strings in `value` appeared in `mapping`.

    Parameters
    ----------
    value : str, iterable or mapping
        Object whose str value needs to be replaced.

    mapping : dict
        Mapping of str value to be replaced and replacement.

    Returns
    -------
    replaced : iterable or mapping
        Value of `value` replaced with `mapping`.
    """
    mapping = mapping.items()

    def replace_with_mapping(s: str):
        """Replace all k in s with v, as in mapping."""
        return reduce(lambda x, kv: x.replace(*kv), mapping, s)

    def replace_dict(d: dict):
        """Replace key and value of a dict."""
        return {replace(k): replace(v) for k, v in d.items()}

    def replace_sequence(it: tuple | list):
        """Replace elements of iterable."""
        return type(it)(map(replace, it))

    def replace(v):
        """The main replace function."""
        if isinstance(v, str):
            return replace_with_mapping(v)
        elif isinstance(v, list | tuple):
            return replace_sequence(v)
        elif isinstance(v, dict):
            return replace_dict(v)
        else:
            return v

    return replace(value)


def report_interval(
    vmid: float,
    vmin: float,
    vmax: float,
    precision: int = 2,
    min_exponent: int = 1,
    max_exponent: int = 2,
) -> str:
    r"""Report parameter interval in :math:`\LaTeX` format.

    Parameters
    ----------
    vmid : float
        The mid value.
    vmin : float
        The lower bound.
    vmax : float
        The upper bound.
    precision : int, optional
        The precision of the mid value. The default is 2.
    min_exponent : int, optional
        The minimum exponent to use scientific notation. The default is 1.
    max_exponent : int, optional
        The maximum exponent to use scientific notation. The default is 2.

    Returns
    -------
    str
        The interval in :math:`|LaTeX` format.
    """
    vmid = float(vmid)
    vmin = float(vmin)
    vmax = float(vmax)
    precision = int(precision)
    min_exponent = int(min_exponent)
    max_exponent = int(max_exponent)

    # assert vmin <= vmid <= vmax
    assert precision > 0
    assert min_exponent > 0
    assert max_exponent > 0

    def get_sci_notation_exponent(num: float) -> int:
        """Get the exponent of a number in scientific notation."""
        return int(f'{num:.{precision}e}'.split('e')[1])

    def get_sci_notation_significand(num: float, exp: int) -> str:
        """Get the significand of a number in scientific notation."""
        significand = num * 10**-exp
        if abs(num) < 10 ** (exp - precision):
            p = abs(get_sci_notation_exponent(num) - exp)
            return f'{significand:+.{p}f}'.rstrip('0')
        else:
            p = precision
            return f'{significand:+.{p}f}'

    lower = vmin - vmid
    upper = vmax - vmid
    exponent = math.log10(abs(vmid))

    if exponent <= -min_exponent or exponent >= max_exponent:
        str_mid = f'{vmid:.{precision}e}'.split('e')[0]
        base_exponent = math.floor(exponent)
        suffix = rf' \times 10^{{{base_exponent}}}'
    else:
        str_mid = f'{vmid:.{precision}f}'
        base_exponent = 0
        suffix = ''

    if lower != 0:
        str_lower = get_sci_notation_significand(lower, base_exponent)
    else:
        str_lower = '-0'
    if upper != 0:
        str_upper = get_sci_notation_significand(upper, base_exponent)
    else:
        str_upper = '+0'

    return f'${str_mid}_{{{str_lower}}}^{{{str_upper}}}{suffix}$'


def progress_bar_factory(
    neval: int,
    ncores: int,
    init_str: str | None = None,
    run_str: str | None = None,
    update_rate: int = 50,
) -> Callable[[Callable], Callable]:
    """Add a progress bar to JAX ``fori_loop`` kernel, see [1]_ for details.

    Parameters
    ----------
    neval : int
        The total number of evaluations.
    ncores : int
        The number of cores.
    init_str : str, optional
        The string displayed before progress bar when initialization.
    run_str : str, optional
        The string displayed before progress bar when run.
    update_rate : int, optional
        The update rate of the progress bar. The default is 50.

    Returns
    -------
    progress_bar_fori_loop : callable
        Factory that adds a progress bar to function input.

    References
    ----------
    .. [1] `How to add a progress bar to JAX scans and loops
            <https://www.jeremiecoullon.com/2021/01/29/jax_progress_bar/>`_
    """
    neval = int(neval)
    ncores = int(ncores)
    neval_single = neval // ncores

    if neval % ncores != 0:
        raise ValueError('neval must be multiple of ncores')

    if init_str is None:
        init_str = 'Compiling... '
    else:
        init_str = str(init_str)

    if run_str is None:
        run_str = 'Running'
    else:
        run_str = str(run_str)

    if neval > update_rate:
        print_rate = max(1, int(neval_single / update_rate))
    else:
        print_rate = 1

    # lock serializes access to idx_counter since callbacks are multithreaded
    lock = Lock()
    idx_counter = 0  # resource counter
    remainder = neval_single % print_rate
    bar = tqdm(range(neval))
    bar.set_description(init_str, refresh=True)

    def _update_tqdm(increment):
        bar.set_description(run_str, refresh=False)
        bar.update(int(increment))

    def _close_tqdm():
        nonlocal idx_counter

        bar.update(remainder)

        with lock:
            idx_counter += 1

        if idx_counter == ncores:
            bar.close()

    def _update_progress_bar(iter_num):
        _ = lax.cond(
            iter_num == 1,
            lambda _: io_callback(_update_tqdm, None, 0),
            lambda _: None,
            operand=None,
        )
        _ = lax.cond(
            iter_num % print_rate == 0,
            lambda _: io_callback(_update_tqdm, None, print_rate),
            lambda _: None,
            operand=None,
        )
        _ = lax.cond(
            iter_num == neval_single,
            lambda _: io_callback(_close_tqdm, None),
            lambda _: None,
            operand=None,
        )

    def progress_bar_fori_loop(fn):
        """Decorator that adds a progress bar to `body_fun` used in
        `lax.fori_loop`.
        Note that `body_fun` must be looping over a tuple who's first element
        is `np.arange(num_samples)`.
        This means that `iter_num` is the current iteration number
        """

        def _wrapper_progress_bar(i, vals):
            result = fn(i, vals)
            _update_progress_bar(i + 1)
            return result

        return _wrapper_progress_bar

    return progress_bar_fori_loop


def to_native_byteorder(arr: NDArray) -> NDArray:
    """Convert an array to native byte order."""
    if arr.dtype.byteorder != '=':
        return arr.astype(arr.dtype.newbyteorder('='))
    else:
        return arr



from __future__ import annotations

import jax
from numpyro.infer.ensemble import AIES, ESS
from numpyro.infer.mcmc import MCMCKernel
from numpyro.util import is_prng_key


class NumpyroEnsembleSampler(MCMCKernel):
    """Wrapper kernel to run the ensemble sampler as a single MCMC chain.

    To collect the posterior correctly, get samples in shape of
    (n_parallel, n_steps, n_walkers) by

    .. code-block:: python

        from numpyro.infer import MCMC
        kernel = NumpyroEnsembleSampler(...)
        mcmc = MCMC(kernel, ...)
        mcmc.run(...)
        samples = mcmc.get_samples(group_by_chain=True)

    Then combine the walkers from the same sampler by

    .. code-block:: python

        import jax
        import jax.numpy as jnp
        samples = jax.tree.map(lambda x: jnp.swapaxes(x, 1, 2), samples)
        samples = jax.tree.map(
            lambda x: jnp.reshape(
                x,
                (x.shape[0], x.shape[1] * x.shape[2], *x.shape[3:]),
            ),
            samples,
        )
    """

    _kernel: type[AIES | ESS]

    def __init__(self, walkers: int, *args, **kwargs):
        self._walkers = int(walkers)
        self._sampler = self._kernel(*args, **kwargs)

    @property
    def sample_field(self):
        return 'z'

    def postprocess_fn(self, args, kwargs):
        return jax.vmap(self._sampler.postprocess_fn(args, kwargs))

    def init(self, rng_key, num_warmup, init_params, model_args, model_kwargs):
        if not is_prng_key(rng_key):
            raise NotImplementedError(
                "EnsembleSampler only supports chain_method='parallel' or "
                "chain_method='sequential'. Please put in a feature request "
                'if it would be useful to be used in vectorized mode.'
            )
        rng_keys = jax.random.split(rng_key, self._walkers)
        if init_params is not None:
            if not all(
                param.shape[0] == self._walkers
                for param in jax.tree.leaves(init_params)
            ):
                raise ValueError(
                    'The batch dimension of each param must match chains'
                )

        return self._sampler.init(
            rng_keys, num_warmup, init_params, model_args, model_kwargs
        )

    def sample(self, state, model_args, model_kwargs):
        return self._sampler.sample(state, model_args, model_kwargs)


class NumPyroAIES(NumpyroEnsembleSampler):
    _kernel = AIES

    def get_diagnostics_str(self, state):
        return f'acc. prob={state.inner_state.mean_accept_prob:.2f}'


class NumPyroESS(NumpyroEnsembleSampler):
    _kernel = ESS



"""Miscellaneous plotting."""

from __future__ import annotations

from collections.abc import Sequence

import arviz as az
import corner
import matplotlib.pyplot as plt
import numpy as np

from elisa.plot.util import (
    gaussian_kernel_smooth,
    get_colors,
    get_contour_colors,
)
from elisa.util.misc import report_interval


def plot_corner(
    idata: az.InferenceData,
    bins: int | Sequence[int] = 40,
    hist_bin_factor: float | Sequence[float] = 1.5,
    params: str | Sequence[str] | None = None,
    plot_range: Sequence[float] | None = None,
    axes_scale: str | Sequence[str] = 'linear',
    levels: float | Sequence[float] | None = None,
    titles: str | Sequence[str] | None = None,
    labels: str | Sequence[str] | None = None,
    color: str = None,
    divergences: bool = True,
):
    """Plot posterior corner plot.

    Parameters
    ----------
    idata : az.InferenceData
        arviz.InferenceData object.
    bins : int or list of int, optional
        The number of bins to use in histograms, either as a fixed value for
        all dimensions, or as a list of integers for each dimension.
        The default is 40.
    hist_bin_factor : float or list of float, optional
        This is a factor (or list of factors, one for each dimension) that
        will multiply the bin specifications when making the 1-D histograms.
        This is generally used to increase the number of bins in the 1-D
        plots to provide more resolution. The default is 1.5.
    params : str or list of str, optional
        One or more parameters to be plotted.
    plot_range : float, or list of float, optional
        A list where each parameter is either a length 2 tuple containing
        lower and upper bounds.
    axes_scale : str, or list of str, optional
        Scale to use for each parameter dimension. If only one scale is given,
        use that for all dimensions. Scale must be ``'linear'`` or ``'log'``.
    levels : float, or list of float, optional
        Credible levels to plot. If None, use 0.683 and 0.95.
    titles : str, or list of str, optional
        Titles to be displayed in the diagonal.
    labels : str, or list of str, optional
        Labels to be displayed in axis label.
    color : str, optional
        Color to use for the plot.
    divergences : bool, optional
        Whether to mark diverging samples.

    Returns
    -------
    fig : plt.Figure
        The figure containing the corner plot.

    """
    posterior = idata['posterior']

    if params is None:
        params = list(posterior.data_vars.keys())
    elif isinstance(params, str):
        params = [params]
    else:
        params = list(params)

    posterior = posterior[params]

    all_params = posterior.data_vars.keys()
    not_found = set(params) - set(all_params)
    if not_found:
        raise ValueError(f'parameter {not_found} not found in posterior')

    if titles is None:
        titles = params
    elif isinstance(titles, str):
        titles = [titles]
    else:
        titles = list(titles)
    assert len(titles) == len(params)

    if labels is None:
        labels = params
    elif isinstance(labels, str):
        labels = [labels]
    else:
        labels = list(labels)
    assert len(labels) == len(params)

    median = posterior.median()
    quantile = posterior.quantile([0.15865, 0.84135])
    median = {p: float(median[p].values) for p in params}
    quantile = {p: quantile[p].values.tolist() for p in params}
    titles = [
        f'{t} = {report_interval(median[p], *quantile[p])}'
        for t, p in zip(titles, params, strict=True)
    ]

    if levels is None:
        levels = [
            [0.683, 0.954, 0.997],  # 1/2/3-sigma of 1d
            [0.393, 0.865, 0.989],  # 1/2/3-sigma of 2d
            [0.683, 0.9],  # 1-sigma of 1d, and 90% of 2d
            [0.393, 0.683, 0.9],  # 1-sigma, 68.3% and 90% of 2d
            [0.683, 0.95],  # 68.3% and 95% of 2d
        ][-1]
    else:
        levels = list(levels)

    # def to_hex(c):
    #     rgb_hex = ''.join(f'{round(i * 255):02x}' for i in c[:3])
    #     return f'#{rgb_hex}'
    # cmap = plt.get_cmap('Blues')
    # colors2 = [cmap(i*0.8 + 0.1) for i in levels]
    # colors1 = [scale_color(to_hex(c), 0.95) for c in colors2]

    if color is None:
        color = '#205295'
    else:
        color = str(color)

    plt.rcParams['axes.formatter.min_exponent'] = 3
    c1, c2 = get_contour_colors(color, len(levels), 0.8, 2.0)

    if plot_range is None:
        vmin = {p: posterior[p].values.min() for p in params}
        vmax = {p: posterior[p].values.max() for p in params}
        if any(vmin[p] == vmax[p] for p in params):
            plot_range = [
                (vmin[p], vmax[p]) if vmin[p] != vmax[p] else 0.99
                for p in params
            ]

    fig = corner.corner(
        idata,
        bins=bins,
        range=plot_range,
        axes_scale=axes_scale,
        color=color,
        hist_bin_factor=hist_bin_factor,
        titles=titles,
        labels=labels,
        show_titles=True,
        title_fmt=None,
        quantiles=[0.15865, 0.5, 0.84135],
        use_math_text=True,
        labelpad=-0.08,
        divergences=divergences,
        divergences_kwargs={'color': 'red', 'alpha': 0.3, 'ms': 1},
        var_names=params,
        # kwargs for corner.hist2d
        levels=levels,
        plot_datapoints=True,
        plot_density=False,
        plot_contours=True,
        fill_contours=True,
        no_fill_contours=True,
        contour_kwargs={'colors': c1},
        contourf_kwargs={'colors': ['white'] + c2, 'alpha': 0.75},
        data_kwargs={'color': c2[0], 'alpha': 0.75, 'ms': 1.5},
    )

    return fig


def plot_trace(
    idata: az.InferenceData,
    params: str | Sequence[str] | None = None,
    axes_scale: str | Sequence[str] | None = None,
    labels: str | Sequence[str] | None = None,
) -> plt.Figure:
    """Plot posterior sampling trace.

    Parameters
    ----------
    idata : az.InferenceData
        arviz.InferenceData object.
    params : str, or list of str, optional
        One or more parameters to be plotted.
    axes_scale : str, or list of str, optional
        Scale to use for each parameter dimension. If only one scale is given,
        use that for all dimensions. Scale must be ``'linear'`` or ``'log'``.
    labels : str, or list of str, optional
        Labels to be displayed in y-axis label.

    Returns
    -------
    fig : plt.Figure
        The figure containing the trace plot.
    """
    colors = get_colors(len(idata['posterior']['chain']), palette='bright')
    posterior = idata['posterior']
    deviance = -2.0 * idata['log_likelihood']['total'].values

    if params is None:
        params = list(posterior.data_vars.keys())
    elif isinstance(params, str):
        params = [params]
    else:
        params = list(params)

    all_params = posterior.data_vars.keys()
    not_found = set(params) - set(all_params)
    if not_found:
        raise ValueError(f'parameter {not_found} not found in posterior')

    nparam = len(params)

    if 'sample_stats' in idata:
        idx = idata['sample_stats']['diverging'].values.nonzero()
        diverging_index = idx[1]
        diverging_sample = [
            posterior[p].values[idx[0], idx[1]] for p in params
        ]
    else:
        diverging_index = None
        diverging_sample = None

    if axes_scale is None:
        axes_scale = ['linear'] * nparam
    else:
        if isinstance(axes_scale, str):
            axes_scale = [axes_scale] * nparam
        elif len(axes_scale) != nparam:
            raise ValueError('`axes_scale` must match `params`')

    if labels is None:
        labels = params
    else:
        if isinstance(labels, str):
            labels = [labels]
        else:
            labels = list(labels)

    if len(labels) != nparam:
        raise ValueError('`tex` must match `params`')

    plt.rcParams['axes.formatter.min_exponent'] = 3
    fig, axes = plt.subplots(
        nrows=len(params),
        ncols=3,
        sharey='row',
        squeeze=False,
        gridspec_kw={'width_ratios': [3, 1, 1]},
        figsize=(9, nparam * 2),
        tight_layout=True,
    )
    fig.subplots_adjust(wspace=0.05, hspace=0)
    fig.align_ylabels(axes)

    axes[0, 0].set_title('trace')
    axes[0, 1].set_title('posterior')
    axes[0, 2].set_title('deviance')

    chain = posterior.chain.values
    draw = posterior.draw.values
    ndraw = posterior.draw.size
    bw = max(draw.size // 100, 10)

    for i in range(nparam):
        posterior_i = posterior[params[i]]
        if axes_scale[i] == 'log' and np.all(posterior_i.values > 0):
            scale = 'log'
            log_scale = True
        else:
            scale = 'linear'
            log_scale = False
        axes[i, 0].set_ylabel(labels[i])
        axes[i, 0].set_yscale(scale)
        for c in chain:
            sample = posterior_i[c].values
            loglike = deviance[c]
            draw_slice = draw[:: bw // 2]
            y = np.log(sample) if log_scale else sample
            smoothed = gaussian_kernel_smooth(draw, y, bw, draw_slice)
            x, kde = az.kde(y)
            if log_scale:
                smoothed = np.exp(smoothed)
                x = np.exp(x)

            color = colors[c]
            zorder = 10 - c

            axes[i, 0].step(
                draw + c / chain.size,
                sample,
                c=color,
                alpha=0.4,
                lw=0.15,
                zorder=zorder,
            )
            axes[i, 0].plot(
                draw_slice, smoothed, c=color, alpha=0.6, lw=1.5, zorder=zorder
            )
            axes[i, 1].plot(kde, x, c=color, alpha=0.6, lw=1.5, zorder=zorder)
            axes[i, 2].scatter(
                loglike, sample, color=color, s=0.05, alpha=0.05, zorder=zorder
            )

        axes[i, 0].set_xlim(0.5, ndraw - 0.5)
        # axes[i, 0].set_xticks([])
        axes[i, 1].set_xticks([])

    if diverging_index is not None:
        for i in range(nparam):
            ylim = axes[i, 0].get_ylim()
            y_lower = ylim[0]

            span = 0.1

            if axes_scale[i] == 'linear':
                y_upper = ylim[0] + span * np.diff(ylim)
            else:
                y_upper = ylim[0] * np.exp(span * np.diff(np.log(ylim)))

            axes[i, 0].vlines(diverging_index, y_lower, y_upper, color='k')
            axes[i, 0].set_ylim(ylim)

            xlim = axes[i, 1].get_xlim()
            x_lower = xlim[0]
            x_upper = xlim[0] + span * np.diff(xlim)

            axes[i, 1].hlines(diverging_sample[i], x_lower, x_upper, color='k')
            axes[i, 1].set_xlim(xlim)

    return fig



"""Typing aliases to shorten hints."""

from collections.abc import Callable
from typing import TypeVar

import numpy as np
from jax import Array
from jax.typing import ArrayLike

__all__ = [
    'PyFloat',
    'JAXFloat',
    'Float',
    'PRNGKey',
    'NumPyArray',
    'JAXArray',
    'Array',
    'ArrayLike',
    'CompID',
    'CompName',
    'CompParamName',
    'ParamID',
    'ParamName',
    'NameValMapping',
    'CompIDParamValMapping',
    'CompIDStrMapping',
    'ParamIDStrMapping',
    'ParamIDValMapping',
    'ParamNameValMapping',
    'CompEval',
    'ConvolveEval',
    'ModelEval',
    'ModelCompiledFn',
    'NameLaTeX',
    'AdditiveFn',
]

T = TypeVar('T')

PyFloat = float | np.inexact  # must include 0-d NDArray with float dtype
JAXFloat = Array
Float = PyFloat | JAXFloat

PRNGKey = Array

NumPyArray = np.ndarray
JAXArray = Array
Array = NumPyArray | JAXArray

ArrayLike = ArrayLike

# Type aliases for parameter and model module
CompID = CompName = CompParamName = ParamID = ParamName = str
NameValMapping = dict[CompParamName, JAXFloat]
CompIDParamValMapping = dict[CompID, NameValMapping]
CompIDStrMapping = dict[CompID, str]
ParamIDStrMapping = dict[ParamID, str]
ParamIDValMapping = dict[ParamID, JAXFloat]
ParamNameValMapping = dict[ParamName, JAXFloat]
CompEval = Callable[[JAXArray, NameValMapping], JAXArray]
ConvolveEval = Callable[
    [JAXArray, NameValMapping, Callable[[JAXArray], JAXArray]], JAXArray
]
ModelEval = Callable[[JAXArray, CompIDParamValMapping], JAXArray]
ModelCompiledFn = Callable[[JAXArray, ParamIDValMapping], JAXArray]
NameLaTeX = tuple[str, str]
AdditiveFn = Callable[[JAXArray, ParamIDValMapping], dict[NameLaTeX, JAXArray]]



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from emcee import EnsembleSampler as Sampler, State as EmceeState

from elisa.infer.samplers.ensemble.base import (
    EnsembleSampler,
    EnsembleSamplerState,
)

if TYPE_CHECKING:
    from multiprocessing import Queue

    from numpy.random import Generator


class EmceeSampler(EnsembleSampler):
    def get_sampling_fn(
        self,
        chains: int,
        warmup: int,
        steps: int,
        thinning: int,
        tune: bool | None,
        warmup_kwargs: dict,
        sampling_kwargs: dict,
    ):
        ndim = self._ndim
        if chains is None:
            chains = 4 * ndim
        if tune is None:
            tune = False
        log_prob_fn = self._log_prob_with_blobs
        blobs_dtype = self._blobs_dtype

        def sampling_fn(
            sampler_id: int, state: EnsembleSamplerState, queue: Queue
        ):
            emcee_state = EmceeState(
                coords=state.coords,
                random_state=state.random_state,
            )
            sampler1 = Sampler(
                chains,
                ndim,
                log_prob_fn,
                pool=None,
                args=None,
                kwargs=None,
                vectorize=True,
                blobs_dtype=blobs_dtype,
                parameter_names=None,
                **warmup_kwargs,
            )
            sampler2 = Sampler(
                chains,
                ndim,
                log_prob_fn,
                pool=None,
                args=None,
                kwargs=None,
                vectorize=True,
                blobs_dtype=blobs_dtype,
                parameter_names=None,
                **sampling_kwargs,
            )
            queue.put((sampler_id, 'warmup'))
            for s in sampler1.sample(
                emcee_state,
                iterations=warmup,
                tune=tune,
                store=False,
                progress=False,
            ):
                emcee_state = s
                queue.put((sampler_id, 'update'))
            queue.put((sampler_id, 'sample'))
            for s in sampler2.sample(
                emcee_state,
                iterations=steps,
                tune=tune,
                thin_by=thinning,
                store=True,
                progress=False,
            ):
                emcee_state = s
                queue.put((sampler_id, 'update'))
            queue.put((sampler_id, 'finish'))
            samples = sampler2.get_blobs()
            state = EnsembleSamplerState(
                coords=emcee_state.coords,
                random_state=emcee_state.random_state,
            )
            return samples, state

        return sampling_fn

    def get_random_state(self, seed: int) -> Generator:
        return np.random.default_rng(seed)



from __future__ import annotations

import warnings
from functools import partial
from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp
import numpy as np
from ultranest import ReactiveNestedSampler, read_file

from elisa.infer.samplers.util import ravel_params_names, uniform_reparam_model

if TYPE_CHECKING:
    from collections.abc import Callable

    from numpy.typing import NDArray


class UltraNestSampler:
    def __init__(
        self,
        numpyro_model: Callable,
        model_args: tuple = (),
        model_kwargs: dict | None = None,
        seed: int = 42,
        ignore_nan: bool = False,
        **kwargs: dict,
    ):
        if ignore_nan:
            warnings.warn(
                'setting `ignore_nan` to True may fail to spot potential '
                'issues of the model',
                Warning,
            )

        self._model_info = mi = uniform_reparam_model(
            numpyro_model,
            model_args,
            model_kwargs,
            rng_seed=seed,
        )

        @jax.jit
        @jax.vmap
        def log_prob_fn(cube_and_derived):
            log_p = mi.log_prob_fn(mi.unravel(cube_and_derived[: mi.ndim]))
            if ignore_nan:
                log_p = jnp.nan_to_num(log_p, nan=-1e300)
            return log_p

        self._log_prob_fn = log_prob_fn
        self._sampler = None
        self._sampler_constructor = partial(ReactiveNestedSampler, **kwargs)
        self._seed = seed

    def run(
        self,
        viz_sample_names: list[str] | None = None,
        read_file_config: dict | None = None,
        **kwargs: dict,
    ) -> dict[str, NDArray[float]]:
        mi = self._model_info

        if viz_sample_names is None:
            viz_sample_names = mi.params_names
        else:
            viz_sample_names = list(map(str, viz_sample_names))

        @jax.jit
        @jax.vmap
        def transform(cube):
            samples = mi.postprocess_fn(mi.unravel(cube))
            viz = jnp.hstack([samples[i].ravel() for i in viz_sample_names])
            return jnp.append(cube, viz)

        params_names = []
        for i in mi.params_dtype:
            shape = i[2]
            name = f'u_{i[0]}'
            params_names.extend(ravel_params_names(name, shape))

        samples_dtype = mi.params_dtype + mi.deterministic_dtype
        derived_names = []
        for i in viz_sample_names:
            filtered = list(filter(lambda x: x[0] == i, samples_dtype))
            if any(filtered):
                shape = filtered[0][2]
                derived_names.extend(ravel_params_names(i, shape))

        if read_file_config is None:
            prev_state = np.random.get_state()
            np.random.seed(self._seed)
            sampler = self._sampler = self._sampler_constructor(
                param_names=params_names,
                loglike=lambda x: jax.device_get(self._log_prob_fn(x)),
                transform=lambda x: jax.device_get(transform(x)),
                derived_param_names=derived_names,
                vectorized=True,
            )
            sampler.run(**kwargs)
            np.random.set_state(prev_state)
            u_samples = sampler.results['samples'][:, : mi.ndim]
        else:
            read_file_config = dict(read_file_config)
            read_file_config['x_dim'] = mi.ndim
            sequence, final = read_file(**read_file_config)
            results = sequence | final
            u_samples = results['samples'][:, : mi.ndim]

        u_samples = jax.vmap(mi.unravel)(u_samples)
        samples = jax.vmap(mi.postprocess_fn)(u_samples)
        samples = jax.device_get(samples)
        return samples

    def print_results(self, use_unicode: bool = True):
        if not hasattr(self._sampler, 'results'):
            raise RuntimeError(
                'no results found, please run the sampler first.'
            )
        self._sampler.print_results(use_unicode=use_unicode)

    @property
    def ess(self) -> int:
        if not hasattr(self._sampler, 'results'):
            return 0
        else:
            return int(self._sampler.results['ess'])

    @property
    def lnZ(self) -> tuple[float | None, float | None]:
        if not hasattr(self._sampler, 'results'):
            return None, None
        else:
            return (
                self._sampler.results['logz'],
                self._sampler.results['logzerr'],
            )



"""The parameter."""

from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import TYPE_CHECKING, NamedTuple, get_args

import jax.numpy as jnp
from numpyro.distributions import Distribution, LogUniform, Uniform

from elisa.util.integrate import AdaptQuadMethod, make_integral_factory
from elisa.util.misc import build_namespace

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from typing import Any, Literal

    # from tinygp import kernels, means, noise
    from elisa.util.integrate import IntegralFactory
    from elisa.util.typing import (
        CompID,
        CompParamName,
        JAXArray,
        JAXFloat,
        ParamID,
        ParamIDStrMapping,
        ParamIDValMapping,
    )


class AssignmentTracker:
    """Track component assignment of a parameter."""

    def __init__(self):
        self._history = []

    def append(self, cid: CompID, pname: CompParamName) -> None:
        """Append a new assignment record."""
        record = (cid, pname)
        if record not in self._history:
            self._history.append((cid, pname))

    def remove(self, cid: CompID, pname: CompParamName) -> None:
        """Remove an assignment record."""
        self._history.remove((cid, pname))

    def get_comp_param(
        self,
        comp_ids: Iterable[CompID],
    ) -> tuple[CompID, CompParamName] | None:
        """Get the earliest component assignment record within comp_ids."""
        id_set = set(comp_ids)
        flag = [i[0] in id_set for i in self._history]
        if any(flag):
            return self._history[flag.index(True)]
        else:
            return None


class ParamInfo(NamedTuple):
    """Parameter information."""

    name: str | Callable[[ParamIDStrMapping], str]
    """Plain name of the parameter.

    It is a getter function for composite parameter.
    """

    latex: str | Callable[[ParamIDStrMapping], str]
    r""":math:`\LaTeX` format of the parameter.

    It is a getter function for composite parameter.
    """

    default: Any
    """Default value of the parameter."""

    bound: str
    """Value bound expression of the parameter."""

    prior: str
    """Prior distribution expression of the parameter."""

    log: bool
    """Whether the parameter is parameterized in a logarithmic scale."""

    fixed: bool
    """Whether the parameter is fixed."""

    tracker: AssignmentTracker
    """Component assignment tracker."""

    id_to_value: Callable[[ParamIDValMapping], JAXFloat]
    """Mapping function from id to value."""

    dist: Distribution | None = None
    """NumPyro distribution for the parameter."""

    composite: bool = False
    """Whether the parameter is composite."""

    integrate: IntegralFactory | bool = False
    """Integration factory for the parameter for interval parameter.

    For composite parameter composed by interval parameter, this is True.
    """


class Parameter(ABC):
    """Parameter base."""

    _id: ParamID
    _tracker: AssignmentTracker
    _nodes_id: tuple[ParamID, ...]

    def __init__(self):
        self._id = hex(id(self))[2:]
        self._tracker = AssignmentTracker()
        self._nodes_id = (self._id,)

    def _id_to_label(
        self,
        mapping: ParamIDStrMapping,
        label_type: Literal['name', 'latex'],
    ) -> str:
        """Get the label of the parameter."""
        if label_type not in {'name', 'latex'}:
            raise ValueError(f'unknown label type: {label_type}')

        return mapping[self._id]

    @property
    def _id_to_value(self) -> Callable[[ParamIDValMapping], JAXFloat]:
        """Gets the mapping function from id to value."""
        pid = self._id
        default = self.default

        def id_to_value(mapping: ParamIDValMapping) -> JAXFloat:
            """Get the value of the parameter from mapping."""
            return mapping.get(pid, default)

        return id_to_value

    @property
    @abstractmethod
    def name(self) -> str:
        """Plain name of the parameter."""
        pass

    @property
    @abstractmethod
    def latex(self) -> str:
        r""":math:`\LaTeX` format of the parameter."""
        pass

    @latex.setter
    @abstractmethod
    def latex(self, value):
        pass

    @property
    @abstractmethod
    def default(self) -> JAXFloat:
        """Default value of the parameter."""
        pass

    @default.setter
    @abstractmethod
    def default(self, value):
        pass

    @property
    @abstractmethod
    def log(self) -> bool:
        """Whether the parameter is parameterized in a logarithmic scale."""
        pass

    @log.setter
    @abstractmethod
    def log(self, value):
        pass

    @property
    @abstractmethod
    def fixed(self) -> bool:
        """Whether the parameter is fixed."""
        pass

    @fixed.setter
    @abstractmethod
    def fixed(self, value):
        pass

    @property
    @abstractmethod
    def _info(self) -> dict[ParamID, ParamInfo]:
        """Parameter information."""
        pass

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.name

    def __add__(self, other: Parameter) -> CompositeParameter:
        return self._make_composite_parameter(self, other, '+')

    def __radd__(self, other: Parameter) -> CompositeParameter:
        return self._make_composite_parameter(other, self, '+')

    def __sub__(self, other: Parameter) -> CompositeParameter:
        return self._make_composite_parameter(self, other, '-')

    def __rsub__(self, other: Parameter) -> CompositeParameter:
        return self._make_composite_parameter(other, self, '-')

    def __mul__(self, other: Parameter) -> CompositeParameter:
        return self._make_composite_parameter(self, other, '*')

    def __rmul__(self, other: Parameter) -> CompositeParameter:
        return self._make_composite_parameter(other, self, '*')

    def __truediv__(self, other: Parameter) -> CompositeParameter:
        return self._make_composite_parameter(self, other, '/')

    def __rtruediv__(self, other: Parameter) -> CompositeParameter:
        return self._make_composite_parameter(other, self, '/')

    def __pow__(self, other: Parameter) -> CompositeParameter:
        return self._make_composite_parameter(self, other, '^')

    def __rpow__(self, other: Parameter) -> CompositeParameter:
        return self._make_composite_parameter(other, self, '^')

    @staticmethod
    def _make_composite_parameter(
        lhs: Parameter,
        rhs: Parameter,
        op: Literal['+', '-', '*', '/', '^'],
    ) -> CompositeParameter:
        # check if the type of lhs and rhs are both parameter
        if not (isinstance(lhs, Parameter) and isinstance(rhs, Parameter)):
            raise TypeError(
                f'unsupported operand types for {op}: '
                f"'{type(lhs).__name__}' and '{type(rhs).__name__}'"
            )

        op_symbol = op

        if op == '+':
            op = jnp.add
            op_name = '{} + {}'
            op_latex = '{{{}}} + {{{}}}'
        elif op == '-':
            op = jnp.subtract
            op_name = '{} - {}'
            op_latex = '{{{}}} - {{{}}}'
        elif op == '*':
            op = jnp.multiply
            op_name = '{} * {}'
            op_latex = r'{{{}}} \times {{{}}}'
        elif op == '/':
            op = jnp.divide
            op_name = '{} / {}'
            op_latex = r'{{{}}} / {{{}}}'
        elif op == '^':
            op = jnp.power
            op_name = '{}^{}'
            op_latex = r'{{{}}}^{{{}}}'
        else:
            raise NotImplementedError(f'op {op}')

        return CompositeParameter(
            params=[lhs, rhs],
            op=op,
            op_name=op_name,
            op_latex=op_latex,
            op_symbol=op_symbol,
        )


class ParameterHelper(Parameter):
    """Handle name, latex, and default value of a parameter."""

    def __init__(
        self,
        name: str,
        latex: str | None,
        default: Any | None = None,
    ):
        if latex is None:
            latex = name

        self._name = name
        self.latex = latex

        if default is not None:
            self.default = default

        super().__init__()

    @property
    def name(self) -> str:
        return self._name

    @property
    def latex(self) -> str:
        return self._latex

    @latex.setter
    def latex(self, latex):
        self._latex = str(latex)

    @property
    @abstractmethod
    def default(self) -> Any:
        pass

    @default.setter
    @abstractmethod
    def default(self, value: Any):
        pass


class DistParameter(ParameterHelper):
    r"""Define a parameter by a distribution.

    Parameters
    ----------
    name : str
        Parameter name.
    dist : Distribution
        Numpyro distribution from which the parameter is sampled.
    default : float
        Parameter default value.
    min : float, optional
        Parameter minimum value, for display purpose only. The default is None.
    max : float, optional
        Parameter maximum value, for display purpose only. The default is None.
    log : bool, optional
        Whether the parameter is parameterized in a logarithmic scale.
        The default is False.
    fixed : bool, optional
        Whether the parameter is fixed. The default is False.
    latex : str, optional
        :math:`\LaTeX` format of the parameter. The default is as `name`.
    """

    def __init__(
        self,
        name: str,
        dist: Distribution,
        default: float,
        *,
        min: float | None = None,
        max: float | None = None,
        log: bool = False,
        fixed: bool = False,
        latex: str | None = None,
    ):
        if not isinstance(dist, Distribution):
            raise ValueError('prior must be a numpyro distribution')

        if jnp.shape(default) != ():
            raise ValueError('default must be a scalar')

        if not bool(dist._validate_sample(default)):
            raise ValueError('default should be within the prior support')

        self._dist = dist
        self._default = jnp.asarray(default, float)

        if min is None:
            self._min = None
        else:
            if jnp.shape(min) != ():
                raise ValueError('min must be a scalar')
            self._min = jnp.asarray(min, float)

        if max is None:
            self._max = None
        else:
            if jnp.shape(max) != ():
                raise ValueError('max must be a scalar')
            self._max = jnp.asarray(max, float)

        self._log = bool(log)
        self._fixed = bool(fixed)

        super().__init__(name, latex, default)

    @property
    def default(self) -> JAXFloat:
        return self._default

    @default.setter
    def default(self, default: float):
        if jnp.shape(default) != ():
            raise ValueError('default must be a scalar')

        if not bool(self._dist._validate_sample(default)):
            raise ValueError('default should be within the dist support')

        self._default = jnp.asarray(default, float)

    @property
    def log(self) -> bool:
        return self._log

    @property
    def fixed(self) -> bool:
        return self._fixed

    @property
    def _dist_expr(self) -> str:
        dist = self._dist
        name = dist.__class__.__name__
        args = [
            f'{arg}={getattr(dist, arg):.4g}'
            for arg in self._dist.arg_constraints
        ]
        args = ', '.join(args)
        return f'{name}({args})'

    @property
    def _info(self) -> dict[ParamID, ParamInfo]:
        vmin = f'{self._min:.4g}' if self._min is not None else '???'
        vmax = f'{self._max:.4g}' if self._max is not None else '???'
        if vmin == '???' or vmax == '???':
            bound_expr = str(self._dist.support)
        else:
            bound_expr = f'({vmin}, {vmax})'

        info = ParamInfo(
            name=self.name,
            latex=self.latex,
            default=self.default,
            bound='' if self.fixed else bound_expr,
            prior='' if self.fixed else self._dist_expr,
            log=self.log,
            fixed=self.fixed,
            tracker=self._tracker,
            id_to_value=self._id_to_value,
            dist=None if self.fixed else self._dist,
        )

        return {self._id: info}


class UniformParameter(DistParameter):
    r"""Define the parameter by a uniform distribution.

    Parameters
    ----------
    name : str
        Parameter name.
    default : float
        Parameter default value.
    min : float
        Parameter minimum value.
    max : float
        Parameter maximum value.
    log : bool, optional
        Whether the parameter is logarithmically uniform. The default is False.
    fixed : bool, optional
        Whether the parameter is fixed. The default is False.
    latex : str, optional
        :math:`\LaTeX` format of the parameter. The default is as `name`.
    """

    def __init__(
        self,
        name: str,
        default: float,
        min: float,
        max: float,
        *,
        log: bool = False,
        fixed: bool = False,
        latex: str | None = None,
    ):
        self._log = bool(log)

        self._name = name
        self._check_and_set_values(default, min, max, fixed)

        super().__init__(
            name,
            self._dist,
            default,
            min=min,
            max=max,
            log=log,
            fixed=fixed,
            latex=latex,
        )

    def __repr__(self) -> str:
        if self._fixed:
            return f'{self.name} = {self.default:.4g}'
        elif self._log:
            return f'{self.name} ~ LogUniform({self.min:.4g}, {self.max:.4g})'
        else:
            return f'{self.name} ~ Uniform({self.min:.4g}, {self.max:.4g})'

    @property
    def name(self) -> str:
        return self._name

    @property
    def latex(self) -> str:
        return self._latex

    @latex.setter
    def latex(self, latex: str):
        self._latex = str(latex)

    @property
    def default(self) -> JAXFloat:
        return self._default

    @default.setter
    def default(self, default: float):
        self._check_and_set_values(default=default)

    @property
    def min(self) -> JAXFloat:
        """Parameter minimum value."""
        return self._min

    @min.setter
    def min(self, min: float):
        self._check_and_set_values(min=min)

    @property
    def max(self) -> JAXFloat:
        """Parameter maximum value."""
        return self._max

    @max.setter
    def max(self, max: float):
        self._check_and_set_values(max=max)

    @property
    def log(self) -> bool:
        """Whether the parameter is logarithmically uniform."""
        return self._log

    @log.setter
    def log(self, log: bool):
        log = bool(log)
        if self._log != log:
            self._log = log

            if log:
                self._dist = LogUniform(self._min, self._max)
            else:
                self._dist = Uniform(self._min, self._max)

    @property
    def fixed(self) -> bool:
        return self._fixed

    @fixed.setter
    def fixed(self, fixed: bool):
        self._fixed = bool(fixed)
        self._check_and_set_values(self.default, self.min, self.max)

    @property
    def _dist_expr(self) -> str:
        if self._log:
            return f'LogUniform({self.min:.4g}, {self.max:.4g})'
        else:
            return f'Uniform({self.min:.4g}, {self.max:.4g})'

    def _check_and_set_values(
        self,
        default: float | None = None,
        min: float | None = None,
        max: float | None = None,
        fixed: bool | None = None,
    ) -> None:
        """Check and set parameter configuration."""
        if default is None:
            _default = self._default
        else:
            if jnp.shape(default) != ():
                raise ValueError('default must be a scalar')
            _default = jnp.asarray(default, float)

        if min is None:
            _min = self._min
        else:
            _min = jnp.asarray(float(min), float)

        if max is None:
            _max = self._max
        else:
            _max = jnp.asarray(float(max), float)

        if fixed is None:
            fixed = self._fixed
        else:
            fixed = bool(fixed)

        if _min <= 0.0 and self._log:
            raise ValueError(f'min ({_min}) must be positive for log uniform')

        if _min >= _max:
            raise ValueError(f'min ({_min}) must be less than max ({_max})')

        if _default < _min:
            raise ValueError(
                f'default ({_default}) must be greater than min ({_min})'
            )

        if _default > _max:
            raise ValueError(
                f'default ({_default}) must be less than max ({_max})'
            )

        if _default == _min and not fixed:
            new_default = _min + 1e-10 * (_max - _min)
            warnings.warn(
                f'the default value of {self} ({_default}) is equal '
                f'its min ({_min}), which will lead to undefined result; '
                f'the default value is now reset to {new_default}',
                Warning,
            )
            _default = new_default

        if _default == _max and not fixed:
            new_default = _max - 1e-10 * (_max - _min)
            warnings.warn(
                f'the default value of {self} ({_default}) is equal '
                f'to its max ({_max}), which will lead to undefined result; '
                f'the default value is now reset to {new_default}',
                Warning,
            )
            _default = new_default

        if default is not None:
            self._default = _default

        if min is not None:
            self._min = _min

        if max is not None:
            self._max = _max

        if min is not None or max is not None:
            if self.log:
                self._dist = LogUniform(self._min, self._max)
            else:
                self._dist = Uniform(self._min, self._max)


class ConstantParameter(ParameterHelper):
    r"""Constant parameter.

    Parameters
    ----------
    name : str
        Parameter name.
    value : array_like
        The constant value of parameter.
    latex : str, optional
        :math:`\LaTeX` format of the parameter. The default is as `name`.
    """

    def __init__(self, name: str, value: Any, latex: str | None = None):
        super().__init__(name, latex, value)

    @property
    def log(self) -> bool:
        """Constant parameter is not logarithmically parameterized."""
        return False

    @property
    def fixed(self) -> bool:
        """Constant parameter is fixed."""
        return True


class ConstantValue(ConstantParameter):
    r"""Parameter with a fixed value.

    Parameters
    ----------
    name: str
        Parameter name.
    value: float
        Parameter value.
    latex : str, optional
        :math:`\LaTeX` format of the parameter. The default is as `name`.
    """

    def __init__(self, name: str, value: float, latex: str | None = None):
        super().__init__(name, value, latex)

    def __repr__(self) -> str:
        return f'{self.name} = {self.default:.4g}'

    @property
    def default(self) -> JAXFloat:
        return self._default

    @default.setter
    def default(self, default: float):
        if jnp.shape(default) != ():
            raise ValueError('default must be a scalar')

        self._default = jnp.asarray(default, float)

    @property
    def _info(self) -> dict[ParamID, ParamInfo]:
        info = ParamInfo(
            name=self.name,
            latex=self.latex,
            default=self.default,
            bound='',  # this is not supposed to be used
            prior='',  # this is not supposed to be used
            log=self.log,
            fixed=self.fixed,
            tracker=self._tracker,
            id_to_value=self._id_to_value,
        )

        return {self._id: info}


class ConstantInterval(ConstantParameter):
    r"""Constant parameter to be integrated over a given interval.

    When assigning :class:`ConstantInterval` parameters to a model component,
    the model will be evaluated according to the following formula:

    .. math::
        \frac{1}{\prod_i (b_i - a_i)}
        \int f(E, \vec{\theta}(\vec{p}, \vec{q})) \, \mathrm{d} \vec{p}

    where :math:`f` is the model function, :math:`\vec{\theta}` is the
    parameter vector of the model, :math:`\vec{p}` is the
    :class:`ConstantInterval` parameters, :math:`\vec{q}` is the other
    parameters, and :math:`a_i` and :math:`b_i` are the intervals given
    by :math:`\vec{p}`.

    Parameters
    ----------
    name: str
        Parameter name.
    interval: array_like
        The interval, a 2-element sequence.
    method : {'quadgk', 'quadcc', 'quadts', 'romberg', 'rombergts'}, optional
        Numerical integration method used to integrate over the parameter.
        Available options are:

            * ``'quadgk'``: global adaptive quadrature by Gauss-Konrod rule
            * ``'quadcc'``: global adaptive quadrature by Clenshaw-Curtis rule
            * ``'quadts'``: global adaptive quadrature by trapz tanh-sinh rule
            * ``'romberg'``: Romberg integration
            * ``'rombergts'``: Romberg integration by tanh-sinh
              (a.k.a. double exponential) transformation

        The default is ``'quadgk'``.
    latex : str, optional
        :math:`\LaTeX` format of the parameter. The default is as `name`.
    kwargs : dict, optional
        Extra kwargs passed to integration methods. See [1]_ for details.

    References
    ----------
    .. [1] `quadax docs <https://quadax.readthedocs.io/en/latest/api.html#adaptive-integration-of-a-callable-function-or-method>`__
    """

    def __init__(
        self,
        name: str,
        interval: Sequence[float],
        method: AdaptQuadMethod = 'quadgk',
        latex: str | None = None,
        **kwargs,
    ):
        super().__init__(name, interval, latex)
        self.method = method
        self._integrate_kwargs = {'epsabs': 0.0, 'epsrel': 1.4e-8} | kwargs

    def __repr__(self) -> str:
        return f'{self.name} = [{self.default[0]:.4g}, {self.default[1]:.4g}]'

    @property
    def default(self) -> JAXArray:
        return self._default

    @default.setter
    def default(self, default):
        if jnp.shape(default) != (2,):
            raise ValueError('interval must be a 2-element sequence')

        self._default = jnp.asarray(default, float)

    @property
    def method(self) -> AdaptQuadMethod:
        """Numerical integration method."""
        return self._method

    @method.setter
    def method(self, value: AdaptQuadMethod):
        supported = get_args(AdaptQuadMethod)
        if value not in supported:
            raise ValueError(f'method must be one of {supported}')

        self._method = value

    @property
    def _info(self) -> dict[ParamID, ParamInfo]:
        factory = make_integral_factory(
            param_id=self._id,
            interval=self.default,
            method=self.method,
            kwargs=self._integrate_kwargs,
        )

        info = ParamInfo(
            name=self.name,
            latex=self.latex,
            default=self.default,
            bound='',  # this is not supposed to be used
            prior='',  # this is not supposed to be used
            log=self.log,
            fixed=self.fixed,
            tracker=self._tracker,
            integrate=factory,
            id_to_value=self._id_to_value,
        )

        return {self._id: info}


# class DependentInterval(Parameter):
#     """Interval whose bounds are defined by function of parameters."""


class CompositeParameter(Parameter):
    r"""Combine parameters to create a new parameter.

    Parameters
    ----------
    params : Parameter, or sequence of Parameter
        Parameters to be combined.
    op : callable
        Function applied to `params`. The function should take the same
        number and order of arguments as `params` and return a single value.
        The function must be compatible with :mod:`JAX`.
    op_name : str
        Name of the composition operator `op`.
    op_latex : str, optional
        :math:`\LaTeX` format of the composition operator `op`.
        The default is as `op_name`.
    """

    _params: tuple[Parameter, ...]
    _has_interval: bool

    def __init__(
        self,
        params: Parameter | Sequence[Parameter],
        op: Callable[..., JAXFloat],
        op_name: str,
        op_latex: str | None = None,
        *,
        op_symbol: Literal['+', '-', '*', '/', '^'] | None = None,
    ):
        # check if the type of params is parameter or sequence
        if not isinstance(params, Parameter | Sequence):
            raise TypeError(
                'parameters must be a Parameter or a sequence of Parameter'
            )

        # make params a list
        if isinstance(params, Parameter):
            params = [params]
        else:
            params = list(params)

        # check if params are all parameters
        if not all(isinstance(i, Parameter) for i in params):
            raise TypeError(
                'parameters must be a Parameter or a sequence of Parameter'
            )

        if op_symbol not in {'+', '-', '*', '/', '^', None}:
            raise ValueError('`op_symbol` is for internal use only')

        super().__init__()

        self._op = op
        self._op_name = str(op_name)
        self._op_latex = self._op_name if op_latex is None else str(op_latex)
        self._op_symbol = op_symbol
        self._params = tuple(params)
        self._latex_custom = None

        for p in self._params:
            if isinstance(p, ConstantInterval) or (
                isinstance(p, CompositeParameter) and p._has_interval
            ):
                self._has_interval = True
                break
        else:
            self._has_interval = False

        # correct Parameter's _nodes_id attribute
        nodes = []
        for p in self._params:
            stack = [p]
            while stack:
                node = stack.pop(0)
                if isinstance(node, CompositeParameter):
                    stack = list(node._params) + stack
                elif node not in nodes:
                    nodes.append(node)
        self._nodes = tuple(nodes)
        self._nodes_id = tuple(p._id for p in self._nodes)

        pid_to_pname = dict(
            zip(
                self._nodes_id,
                build_namespace([p.name for p in self._nodes], prime=True)[
                    'namespace'
                ],
                strict=True,
            )
        )
        self._name = self._id_to_label(pid_to_pname, 'name')

    def _id_to_label(
        self,
        mapping: dict[ParamID, str],
        label_type: Literal['name', 'latex'],
    ) -> str:
        if label_type not in {'name', 'latex'}:
            raise ValueError(f'unknown label type: {label_type}')

        if self._id in mapping:
            return mapping[self._id]
        else:
            if self._op_symbol:
                op = self._op_symbol
                lhs, rhs = self._params

                if op == '+':
                    lhs_fmt = rhs_fmt = '{}'

                elif op == '-':
                    lhs_fmt = '{}'

                    if isinstance(
                        rhs, CompositeParameter
                    ) and rhs._op_symbol not in {'*', '/', '^'}:
                        rhs_fmt = '({})'
                    else:
                        rhs_fmt = '{}'

                elif op == '*':
                    if isinstance(
                        lhs, CompositeParameter
                    ) and lhs._op_symbol not in {'*', '/', '^'}:
                        lhs_fmt = '({})'
                    else:
                        lhs_fmt = '{}'

                    if isinstance(
                        rhs, CompositeParameter
                    ) and rhs._op_symbol not in {'*', '/', '^'}:
                        rhs_fmt = '({})'
                    else:
                        rhs_fmt = '{}'

                elif op == '/':
                    if isinstance(
                        lhs, CompositeParameter
                    ) and lhs._op_symbol not in {'*', '/', '^'}:
                        lhs_fmt = '({})'
                    else:
                        lhs_fmt = '{}'

                    rhs_fmt = '({})'

                elif op == '^':
                    if isinstance(lhs, CompositeParameter):
                        lhs_fmt = '({})'
                    else:
                        lhs_fmt = '{}'

                    if isinstance(rhs, CompositeParameter):
                        rhs_fmt = '({})'
                    else:
                        rhs_fmt = '{}'

                else:
                    raise NotImplementedError(f'op_symbol: {op}')

                labels = (
                    lhs_fmt.format(lhs._id_to_label(mapping, label_type)),
                    rhs_fmt.format(rhs._id_to_label(mapping, label_type)),
                )

            else:
                labels = tuple(
                    (
                        '({})' if isinstance(p, CompositeParameter) else '{}'
                    ).format(p._id_to_label(mapping, label_type))
                    for p in self._params
                )

            temp = self._op_name if label_type == 'name' else self._op_latex
            return temp.format(*labels)

    @property
    def _id_to_value(self) -> Callable[[ParamIDValMapping], JAXFloat]:
        fns = [param._id_to_value for param in self._params]

        def id_to_value(mapping: ParamIDValMapping) -> JAXFloat:
            """Get the value of the composite parameter."""
            return self._op(*[fn(mapping) for fn in fns])

        return id_to_value

    @property
    def name(self) -> str:
        return self._name

    @property
    def latex(self) -> str:
        if self._latex_custom is not None:
            return self._latex_custom
        nodes_latex = [p.latex for p in self._nodes]
        latex = build_namespace(nodes_latex, True, True)['namespace']
        pid_to_latex = dict(zip(self._nodes_id, latex, strict=True))
        return self._id_to_label(pid_to_latex, 'latex')

    @latex.setter
    def latex(self, value: str | None):
        if value is None:
            self._latex_custom = None
        else:
            self._latex_custom = str(value)

    @property
    def default(self) -> JAXFloat:
        if self._has_interval:
            raise RuntimeError(
                'cannot get default value of a composite interval'
            )

        return self._op(*[i.default for i in self._params])

    @property
    def log(self) -> bool:
        """If the sub-parameters are all logarithmically parameterized."""
        return all(i.log for i in self._params)

    @property
    def fixed(self) -> bool:
        """If the sub-parameters are all fixed."""
        return all(i.fixed for i in self._params)

    @property
    def _info(self) -> dict[ParamID, ParamInfo]:
        info = {
            self._id: ParamInfo(
                name=lambda mapping: self._id_to_label(mapping, 'name'),
                latex=lambda mapping: self._id_to_label(mapping, 'latex'),
                default=jnp.nan,  # this is not supposed to be used
                bound='',  # this is not supposed to be used
                prior='',  # this is not supposed to be used
                log=self.log,
                fixed=self.fixed,
                tracker=self._tracker,
                id_to_value=self._id_to_value,
                integrate=self._has_interval,
                composite=True,
            )
        }

        for p in self._params:
            info |= p._info

        return info


# class GPParameter(ParameterHelper):
#     """Parameter sampled from a Gaussian process."""
#
#     def __init__(
#         self,
#         name: str,
#         kernel: kernels.Kernel,
#         x: JAXFloat,
#         *,
#         diag: JAXFloat | None = None,
#         noise: noise.Noise | None = None,
#         mean: means.MeanBase | Callable | JAXFloat | None = None,
#         log: bool = False,
#         latex: str | None = None,
#     ):
#         self._log = bool(log)
#         super().__init__(name, latex, None)
#
#     def _fn(self, name: str | None, rng_key: PRNGKey | None) -> JAXFloat:
#         """Sample from the Gaussian process."""
#         raise NotImplementedError
#
#     @property
#     def default(self) -> None:
#         return None
#
#     @property
#     def log(self) -> bool:
#         return self._log
#
#     @property
#     def fixed(self) -> bool:
#         return False
#
#     @property
#     def _info(self) -> dict[ParamID, ParamInfo]:
#         return ...



from .base import (
    ObservationData as ObservationData,
    ResponseData as ResponseData,
    SpectrumData as SpectrumData,
)
from .ogip import (
    Data as Data,
    Response as Response,
    Spectrum as Spectrum,
)



from .add import *  # noqa: F403
from .conv import *  # noqa: F403
from .model import (
    AnaIntAdditive as AnaIntAdditive,
    AnaIntMultiplicative as AnaIntMultiplicative,
    ConvolutionComponent as ConvolutionComponent,
    NumIntAdditive as NumIntAdditive,
    NumIntMultiplicative as NumIntMultiplicative,
    ParamConfig as ParamConfig,
    PyAnaInt as PyAnaInt,
    PyNumInt as PyNumInt,
)
from .mul import *  # noqa: F403
from .parameter import (
    CompositeParameter as CompositeParameter,
    ConstantInterval as ConstantInterval,
    ConstantValue as ConstantValue,
    DistParameter as DistParameter,
    UniformParameter as UniformParameter,
)
